
import yaml, sys, os, random
import pandas as pd
import numpy as np
import subprocess
import json
from dreem.aggregate import poisson
import dreem.util.util as util
from dreem.util.util import *
from dreem.util.dump import *
import jsbeautifier
from dreem.aggregate.rnastructure import RNAstructure


test_files_dir = os.getcwd()+'/test_output' #os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), '../..','test_output'))
input_dir = os.path.join(test_files_dir,'input')
prediction_dir = os.path.join(test_files_dir,'expected_output')
output_dir = os.path.join(test_files_dir,'output')

def hamming_distance(s1, s2):
    return sum(c1 != c2 for c1, c2 in zip(s1, s2))

def generate_barcodes(barcode_length, n, min_barcode_distance):
    barcodes = []
    while(len(barcodes) < n):
        barcode = ''.join(random.choice('ATCG') for _ in range(barcode_length))
        if all(hamming_distance(barcode, b) > min_barcode_distance for b in barcodes):
            barcodes.append(barcode)
    return barcodes

def invert_sequence(seq):
    return ''.join([{'A':'T','T':'A','C':'G','G':'C'}[s] for s in seq])[::-1]

def next_base(base):
    return {'A':'C','C':'G','G':'T','T':'A',0:1}[base]

def prev_base(base):
    return {'A':'T','C':'A','G':'C','T':'G',0:1}[base]

def create_sequence(length, bases=['A','T','C','G']):
    return ''.join([random.choice(bases) for _ in range(length)])

def copy_prediction_as_results(path_expected, path_output):
    assert os.path.exists(path_expected), 'The expected output folder doesn\'t exist'
    os.makedirs(path_output, exist_ok=True)
    os.system('cp -r {} {}'.format(path_expected, path_output))
    
def print_fasta_line(f, id, seq):
    f.write('>{}\n{}\n'.format(id, seq))

def print_fastq_line(f, id, seq, qual):
    f.write('@{}\n{}\n+\n{}\n'.format(id, seq, qual))    

def generate_fastq_files(path, sample_profile, construct=None, max_barcode_muts=None):
    """Write a fastq file with the given parameters
    
    Arguments:
        path {str} -- where to write the fastq files
        sample_profile {dict} -- dictionary with the following keys
            'constructs' -- list of construct names [list]
                'reference' -- sequence to use for each construct [str]
                'number_of_reads' -- number of reads to generate for each construct [list]
                'mutations' -- number of mutations to introduce in each read [list]
                'deletions' -- number of deletions to introduce in each read [list]
                'insertions' -- number of insertions to introduce in each read [list]
        construct {str} -- if specified, only generate fastq files for this construct
    """
    if construct is not None:
        f_prefix = construct
    else:
        f_prefix = path.split('/')[-1]
    fastq1_name, fastq2_name = os.path.join(path, f_prefix + '_R1.fastq'), os.path.join(path, f_prefix + '_R2.fastq')
    with open(fastq1_name, 'w') as f1, open(fastq2_name, 'w') as f2:
        # write placeholder reads
        for c, v in sample_profile.items():
            for i in range(v['number_of_reads']):
                if max_barcode_muts is not None:
                    bs, be = v['barcode_start'], v['barcode_start'] + len(v['barcodes'])
                    if len([m for m in v['mutations'][i] if m >= bs and m < be]) > max_barcode_muts:
                        continue
                sequence = sample_profile[c]['reads'][i]

                cigar = make_cigar(len(v['reference']), v['mutations'][i], v['insertions'][i], v['deletions'][i], v['no_info'][i])
                
                #print_fastq_line(f1, '{}:{}:{}'.format(c, i, cigar), sequence, 'F'*len(v['reference']))
                #print_fastq_line(f2, '{}:{}:{}'.format(c, i, cigar), invert_sequence(sequence), 'F'*len(v['reference']))
                print_fastq_line(f1, '{}:{}'.format(c, i), sequence, 'F'*len(v['reference']))
                print_fastq_line(f2, '{}:{}'.format(c, i), invert_sequence(sequence), 'F'*len(v['reference']))
    
def generate_fasta_file(filename, sample_profile):
    """Write a fasta file with the given parameters
    
    Arguments:
        filename {str} -- where to write the fasta file
        sample_profile {dict} -- dictionary with the following keys
            'constructs' -- list of construct names [list]
                'reads' -- sequence to use for each construct [list]
    """
    with open(filename, 'w') as f:
        for c, v in sample_profile.items():
            print_fasta_line(f, c, v['reference'])
            


def sam_to_bam(sam_file, bam_file):
    os.system(' '.join(["samtools", "view", "-u",sam_file,'-o', bam_file])   )
    
def bam_to_sam(bam_file, sam_file):
    os.system(' '.join(["samtools", "view", '-h', bam_file,'-o', sam_file]))
      
def generate_library_file(filename, sample_profile):
    """Write a library file with the given parameters

    Arguments:
        filename {str} -- where to write the library file
        sample_profile {dict} -- dictionary with the following keys
            constructs {list} -- list of construct names [list]
                barcodes {list} -- list of barcodes to use for each construct [list]
                barcode_start {int} -- where to start the barcode in the read [int]
    """

    df = pd.DataFrame(sample_profile).T.reset_index().rename(columns={'index':'construct'})
    content_cols = ['construct']
    if 'barcodes' in df.columns:
        df.rename(columns={'barcodes':'barcode'}, inplace=True)
        content_cols += ['barcode_start','barcode']
    if 'sections' in df.columns:
        df.rename(columns={'sections':'section'}, inplace=True)
        content_cols += ['section','section_start','section_end']
        df = df.explode(['section','section_start','section_end'])[content_cols]
        df['section_start'] = df['section_start'].apply(lambda x: x+1) # convert to 1-based
    df['some_random_attribute'] = 'some_random_value'
    df.to_csv(filename, index=False)

def generate_demultiplexed_fastq_files(folder, sample_profile):
    """Demultiplex a fastq file based on a library file

    Arguments:
        folder {str} -- where to write the demultiplexed fastq files
        sample_profile {dict} -- dictionary with the following keys
            'constructs' -- list of construct names [list]
                'reads' -- sequence to use for each construct [list]
                'number_of_reads' -- number of reads to generate for each construct [list]
                'mutations' -- number of mutations to introduce in each read [list]
                'deletions' -- number of deletions to introduce in each read [list]
                'insertions' -- number of insertions to introduce in each read [list]
    """

    for c, v in sample_profile.items():
        generate_fastq_files(folder, {c:v}, c, max_barcode_muts=1)


def print_sam_header(f, construct, len_sequence):
    f.write('@HD\tVN:1.6\tSO:coordinate\n' + '@SQ\tSN:{}\tLN:{}\n'.format(construct, len_sequence) + '@PG	ID:bowtie2	PN:bowtie2	VN:2.4.5\n')

def print_sam_lines(f, read_name, sequence, construct, cigar, sep='\t'):
    f.write(read_name + sep + '99' + sep + construct + sep + '1' + sep + '255' + sep + cigar + sep + '*' + sep + '0' + sep + '0' + sep + sequence + sep + 'F'*len(sequence)+ sep +'\n')
    f.write(read_name + sep + '147' + sep + construct + sep + '1' + sep + '255' + sep + cigar + sep + '*' + sep + '0' + sep + '0' + sep + sequence + sep + 'F'*len(sequence)+ sep +'\n')
    

def make_cigar(len_sequence, mutations, insertions, deletions, no_info):
    """Create a cigar string for a read with the given mutations and indels
    The sequence is assumed to be 1-based.
    The cigar string is 1-based.
    The mutations and indels are 0-based.

    The mutations and indels are assumed to be sorted. This is checked at the beginning of the code.

    Arguments:
        len_sequence {int} -- length of the sequence [int]
        mutations {list} -- list of mutations [list]
        insertions {list} -- list of insertions [list]
        deletions {list} -- list of deletions [list]
    """

    # check that the mutations and indels are sorted
    assert sorted(mutations) == mutations
    assert sorted(insertions) == insertions
    assert sorted(deletions) == deletions
    assert sorted(no_info) == no_info

    for i in range(len(mutations)-1):
        assert mutations[i]+1 != mutations[i+1], 'Consecutive mutations are not allowed'
    
    for i in range(len(insertions)-1):
        assert insertions[i]+1 != insertions[i+1], 'Consecutive insertions are not allowed'
    
    for i in range(len(deletions)-1):
        assert deletions[i]+1 != deletions[i+1], 'Consecutive deletions are not allowed'
        
    for i in range(len(no_info)-1):
        assert no_info[i]+1 != no_info[i+1], 'Consecutive no_info are not allowed'

    # initialize the cigar string
    cigar = ''

    # keep track of the current position in the sequence
    current_position = 1

    # keep track of the current position in the cigar string
    current_cigar_position = 0
    
    # loop through the mutations, insertions, and deletions to create the cigar string. If there's no mutation, insertion, or deletion at a position, then the position is a match.
    while current_position <= len_sequence:
        # if there's a mutation at the current position, then add a mutation to the cigar string
        if current_position-1 in mutations:
            if current_cigar_position > 0:
                cigar += '{}='.format(current_cigar_position)
            cigar += '1X'
            current_cigar_position = 0
            current_position += 1
        # if there's an insertion at the current position, then add an insertion to the cigar string
        elif current_position-1 in insertions:
            if current_cigar_position > 0:
                cigar += '{}='.format(current_cigar_position)
            cigar += '1I'
            current_cigar_position = 1
            current_position += 1
            len_sequence -=1
        # if there's a deletion at the current position, then add a deletion to the cigar string
        elif current_position-1 in deletions:
            if current_cigar_position > 0:
                cigar += '{}='.format(current_cigar_position)
            cigar += '1D'
            current_cigar_position = 0
            current_position += 1
        elif current_position-1 in no_info:
            if current_cigar_position > 0:
                cigar += '{}='.format(current_cigar_position)
            cigar += '1N'
            current_cigar_position = 0
            current_position += 1
        # if there's no mutation, insertion, or deletion at the current position, then add a match to the cigar string
        else:
            current_cigar_position += 1
            current_position += 1
    
    # add the last match to the cigar string
    cigar += '{}='.format(current_cigar_position)
    
    # add softclipping
    soft_clipping = len(deletions) - len(insertions)
    if soft_clipping > 0:
        cigar = cigar + '{}S'.format(soft_clipping)
        
    #cigar = cigar + str(insertions) + str(deletions) 

    return cigar


def generate_sam_files(folder, sample_profile):
    """Generate sam files for the given parameters. 
    The sam files are written from the input parameters, not from an external software. They do not use fastq files.

    Arguments:
        folder {str} -- where to write the sam files
        sample_profile {dict} -- dictionary of parameters for each construct
            - constructs {list} -- list of construct names [list]
                - number_of_reads {list} -- number of reads to generate for each construct [list]
                - mutations {list} -- number of mutations to introduce in each read for each construct [list(list)]
                - reads {list} -- sequence to use for each construct [list]
                - insertions {list} -- number of insertions to introduce in each read for each construct [list(list)]
                - deletions {list} -- number of deletions to introduce in each read for each construct [list(list)]
                - barcode_start {int} -- where to start the barcode in the read [int]
                - barcodes {list} -- list of barcodes to use for each construct [list]            
    """

    for c, v in sample_profile.items():
        with open(os.path.join(folder, '{}.sam'.format(c)), 'w') as f:
            print_sam_header(f, c, len(v['reference']))
            for i in range(v['number_of_reads']):
                cigar = make_cigar(len( v['reference']), v['mutations'][i], v['insertions'][i], v['deletions'][i], v['no_info'][i])
                #print_sam_lines(f, '{}:{}:{}'.format(c, i, cigar), v['reads'][i], c, cigar)
                print_sam_lines(f, '{}:{}'.format(c, i), v['reads'][i], c, cigar)
    
def generate_bam_files(folder, sample_profile):
    for c in sample_profile.keys():
        sam_to_bam(os.path.join(folder, '{}.sam'.format(c)), os.path.join(folder, '{}.bam'.format(c)))


def check_sample_profile(reads, mutations, insertions, deletions):
    """ Check that the sample profile is valid.
    Make sure that:
    - the mutations, insertions, and deletions are sorted
    - the mutations, insertions, and deletions are not consecutive
    - the mutations, insertions, and deletions are not at the same position
    - the mutations, insertions, and deletions are within the sequence length    
    """
    for idx, s in enumerate(reads):
        [mutations[idx][i].sort() for i in range(len(mutations[idx])) if len(mutations[idx])]
        [insertions[idx][i].sort() for i in range(len(insertions[idx])) if len(insertions[idx])]
        [deletions[idx][i].sort() for i in range(len(deletions[idx])) if len(deletions[idx])]
        for j in range(len(mutations[idx])):
            if len(mutations[idx][j]):
                for i in range(len(mutations[idx][j])):
                    assert mutations[idx][j][i] < len(s[j]), 'Mutation is out of sequence length'
            if len(insertions[idx][j]):
                for i in range(len(insertions[idx][j])):
                    assert insertions[idx][j][i] < len(s[j]), 'Insertion is out of sequence length'
            if len(deletions[idx][j]):
                for i in range(len(deletions[idx][j])):
                    assert deletions[idx][j][i] < len(s[j]), 'Deletion is out of sequence length'
        

def make_sample_profile(constructs, reads, number_of_reads, mutations, insertions, deletions, no_info, barcodes=None, barcode_start=None, sections=None, section_start=None, section_end=None):
    """Make a dictionary of the sample profile.
    The dictionary is used to make sure that the sample profile is valid. 
    The lists are grouped by construct, and the number of reads is grouped by construct.
    """
    check_sample_profile(reads, mutations, insertions, deletions)
    if barcodes is not None:
        assert len(constructs) == len(barcodes)
        for b in barcodes:
            assert len(b) == len(barcodes[0])
    
    sample_profile = {}
               
    for idx, c in enumerate(constructs):
        sample_profile[c] = {}
        sample_profile[c]['reads'] = reads[idx]            
        if barcodes is not None:
            for j in range(len(sample_profile[c]['reads'])): 
                sample_profile[c]['reads'][j] = sample_profile[c]['reads'][j][:barcode_start] + barcodes[idx] + sample_profile[c]['reads'][j][barcode_start+len(barcodes[idx]):]
        sample_profile[c]['reference'] = reads[idx][0] 
        sample_profile[c]['number_of_reads'] = number_of_reads[idx]
        sample_profile[c]['mutations'] = mutations[idx]
        sample_profile[c]['insertions'] = insertions[idx]
        sample_profile[c]['deletions'] = deletions[idx]
        sample_profile[c]['no_info'] = no_info[idx]
        if sections is not None:
            sample_profile[c]['sections'] = sections[idx]
            sample_profile[c]['section_start'] = section_start[idx]
            sample_profile[c]['section_end'] = section_end[idx]
            # assert that the sections are within the sequence length
            for j in range(len(sample_profile[c]['sections'])):
                assert sample_profile[c]['section_start'][j] <= len(sample_profile[c]['reference']) -1, 'Section start is out of sequence length'
                assert sample_profile[c]['section_end'][j] <= len(sample_profile[c]['reference']), 'Section end is out of sequence length'
        if barcodes is not None:
            sample_profile[c]['barcodes'] = barcodes[idx]
            sample_profile[c]['barcode_start'] = barcode_start # +1 # 1-based #TODO
        for i in range(sample_profile[c]['number_of_reads']):
            sequence = sample_profile[c]['reads'][i]
            for j in sample_profile[c]['mutations'][i]:
                sequence = sequence[:j] + next_base(sequence[j]) + sequence[j+1:]
            for j in sample_profile[c]['insertions'][i]:
                j += len([k for k in sample_profile[c]['insertions'][i] if k < j])
                sequence = sequence[:j] + next_base(sequence[j]) + sequence[j:]
            for j in sample_profile[c]['deletions'][i]:
                j += len([k for k in sample_profile[c]['insertions'][i] if k < j])
                j -= len([k for k in sample_profile[c]['deletions'][i] if k < j])
                sequence = sequence[:j] + sequence[j+1:]
            sample_profile[c]['reads'][i] = sequence[0:len(sample_profile[c]['reference'])] + ''.join([random.choice(['A', 'C', 'G', 'T']) for _ in range(len(sample_profile[c]['reference']) - len(sequence))])
        #sample_profile[c]['mutations'] = [[mm+1 for mm in m] for m in sample_profile[c]['mutations']]
        #sample_profile[c]['insertions'] = [[ii+1 for ii in i] for i in sample_profile[c]['insertions']]
        #sample_profile[c]['deletions'] = [[dd+1 for dd in d] for d in sample_profile[c]['deletions']]
        #sample_profile[c]['no_info'] = [[nn+1 for nn in n] for n in sample_profile[c]['no_info']]
    return sample_profile






def generate_bitvector_files(folder, sample_profile, library):
    """Generate a bitvector file for the sample profile. 
    
    The bitvector is an orc file generated using pandas. A bitvector is a binary file that contains the reads for a certain section of a certain construct. The sections of the construct are defined by the library. 
    For each section of each construct, we are interested in the residues inside the [section_start, section_end] 0-indexed interval. 
    The bitvector columns are the residues, and the rows are the reads. The columns are under the format '0-indexed position'+'residue'. Ex: '0A' means that the first residue is an A.

    Parameters
    ----------
    folder : str
        The folder to write the bitvector file to.
    sample_profile : dict
        The sample dictionary.
        Attributes are:
        - constructs
            - reads
            - number_of_reads
            - mutations
            - insertions
            - deletions
    library : str
        The library to use. Columns are 'construct', 'section', 'section_start','section_end'.

    """

    library = pd.read_csv(library)

    # get through the constructs
    for construct in sample_profile:

    # create a folder under the construct's name
        construct_folder = os.path.join(folder, construct)
        os.makedirs(construct_folder, exist_ok=True)
    # get through the sections for that construct
        for section in library[library['construct'] == construct]['section']:
            section_folder = os.path.join(construct_folder, section)
            os.makedirs(section_folder, exist_ok=True)
            section_start = library[(library['construct'] == construct) & (library['section'] == section)]['section_start'].values[0] -1 # 0-based
            section_end = library[(library['construct'] == construct) & (library['section'] == section)]['section_end'].values[0]
            sequence = sample_profile[construct]['reference'][section_start:section_end]
            assert len(sequence) == section_end - section_start, 'Section length is not equal to the length of the sequence'
            columns = [sequence[i] + str(i+1 + section_start) for i in range(section_end - section_start)]

            # create a dataframe with the columns
            df = pd.DataFrame(columns=columns, index = list(range(sample_profile[construct]['number_of_reads'])))
                
            for read_number in range(sample_profile[construct]['number_of_reads']):
                df.iloc[read_number] = update_read(sample_profile[construct], [section_start, section_end], read_number)

            df = pd.DataFrame(data=np.array(df, dtype=np.byte), columns=df.columns, index = list(range(sample_profile[construct]['number_of_reads'])))
            
            df.to_orc(os.path.join(section_folder, '0.orc'), index=False, engine="pyarrow")


def update_read(construct_profile, section, read_number):
    sequence = construct_profile['reference']
    bv = [1]*(section[1]-section[0])
    for indexes in group_sequence_by_base_idx(sequence[section[0]:section[1]], section[0]):
        for attribute in ['insertions', 'deletions']:
            for event in construct_profile[attribute][read_number]:
                single_base = len(indexes) ==  1
                if attribute == 'insertions' and event > 0:
                    if next_base(sequence[event]) == sequence[event-1]:
                        single_base = False
                        indexes.append(event-1)
                    else:
                        continue
                clear_bit = B_MATCH if single_base else None
                if event in indexes:
                    for idx in indexes:
                        bv[idx] = set_clear_bit(bv[idx], bit_factory(attribute), clear_bit)
                        if attribute == 'insertions' and idx > 0:
                            bv[idx-1] = set_clear_bit(bv[idx-1], B_INS_5)
        
        for event in construct_profile['insertions'][read_number]:
            if event in indexes and event >= section[0] and event < section[1]:
                idx = event-section[0]
                if idx > 0:
                    if next_base(sequence[idx]) == sequence[idx-1]:    
                        continue                      
                bv[idx] = set_clear_bit(bv[idx], bit_factory('insertions'))
                if idx >0:
                    bv[idx-1] = set_clear_bit(bv[idx-1], B_INS_5) 
    
        for event in construct_profile['mutations'][read_number]:
            if event >= section[0] and event < section[1]:
                idx = event-section[0]
                bv[idx] = set_clear_bit(bv[idx], bit_factory('mutations', next_base(sequence[event])), B_MATCH )

        for tail in range(len(construct_profile['insertions'][read_number]) - len(construct_profile['deletions'][read_number])):
            tail_pos = len(sequence) - tail - 1
            if tail_pos >= section[0] and tail_pos < section[1]:
                bv[tail_pos-section[0]] = 0
    
    return bv

def bit_factory(attribute, base=None):
    if attribute == 'insertions':
        return B_INS_3
    if attribute == 'deletions':
        return B_DELET
    if attribute == 'mutations':
        if base == 'A':
            return B_SUB_A
        if base == 'C':
            return B_SUB_C
        if base == 'G':
            return B_SUB_G
        if base == 'T':
            return B_SUB_T
    raise ValueError

def group_sequence_by_base_idx(sequence, offset=0):
    grouped_seq, i, idx_for_this_base = [], 0, []
    while i < len(sequence):
        idx_for_this_base += [i+offset]
        i += 1
        while i < len(sequence):
            if sequence[i-1] == sequence[i]:
                idx_for_this_base += [i+offset]
                i += 1      
            else:
                break          
        grouped_seq.append(idx_for_this_base)
        idx_for_this_base = []         
    return grouped_seq
    
def set_clear_bit(byte, set, clear=None):
    if clear == None:
        return set_bit(byte, set)
    return set_bit(clear_bit(byte, clear), set)

def set_bit(value, bit):
    return value | (1<<bit)

def clear_bit(value, bit):
    return value & ~(1<<bit)

def generate_samples_file(path):
    sample = path.split('/')[-2].split('.')[0]
    assert sample != '', 'Sample name is empty'
    df = pd.DataFrame({
        'sample': [sample],
        'user': ['Napoleon'],
        'date': ['1769-08-15'],
        'exp_env': ['in_vitro'],
        'temperature_k': [310],
        'buffer': ['MgCl2'],
        'DMS_conc_mM': [2],
        'inc_time_tot_secs': [100]
    }, index=[0])
    df.to_csv(path, index=False)

def count_bases(positions, ss, se):
    out = np.zeros(se-ss, dtype=int)
    for p in positions:
        if p >= ss and p < se:
            out[p-ss] += 1
    return out.tolist()

def count_insertions(ins, ref, ss, se):
    out = np.zeros(se-ss, dtype=int)
    for ins_read in ins:
        for indexes in group_sequence_by_base_idx(ref[ss:se], ss):
            for event in ins_read:
                if event > 0:
                    if next_base(ref[event]) == ref[event-1]:
                        indexes.append(event-1)
                        if event in indexes:
                            for idx in indexes:
                                out[idx] += 1
                
                    else:
                        if event in indexes:
                            out[event-ss] += 1
                
    return out.tolist()

def count_substitutions(muts, ref, ss, se):
    out = np.zeros(se-ss, dtype=int)
    for idx, mm in enumerate(muts):
        for m in mm:
            if m >= ss and m < se:
                out[idx-ss] += 1
    return out.tolist()

def count_coverage(num_reads, insertions, deletions, ss, se, ref):
    out = np.ones(len(ref), dtype=int)*num_reads
    for n in range(num_reads):
        tailing = 0
        for _ in insertions[n]:
            tailing += 1
        for _ in deletions[n]:
            tailing -= 1
        for i in range(tailing):
            out[-i-1] -= 1
    return out[ss:se].tolist()

def count_mut_mod(ref, muts, base, ss, se):
    if base == 'N':
        return np.array([np.array([count_bases([b for b in muts[p] if ref[b] == prev_base(bb)], ss, se) for p in range(len(muts))]).sum(axis=0) for bb in ['A','C','T','G']]).sum(axis=0).tolist()
    return np.array([count_bases([b for b in muts[p] if ref[b] == prev_base(base)], ss, se) for p in range(len(muts))]).sum(axis=0).tolist()



def create_reads_clustering(real_structures, mu_unpaired, mu_paired, n_reads, len_seq):
    """Generate reads from real structures.
    
    Inputs:
    -------
    
    real_structures: list of list of int.
        List of the real alternative structures.
    mu_unpaired: list
        Average mutation rate for unpaired bases for each structure.
    mu_paired: float
        Average mutation rate for paired bases.
    n_reads: list of int.
        Amount of reads per structure.
    len_seq: int
        Length of the sequence.
        
    Output:
    -------
    
    reads: dict of list of int.
        The reads as list of residues under the bitvector format.
    pop_avg: list of np.ndarray.
        The population average for each cluster.
    
    """
    
    pop = [np.zeros((n_read, len_seq)) for n_read in n_reads]
    reads = {}
    for i_s, (structure, n_read) in enumerate(zip(real_structures, n_reads)):
        for i_r in range(n_read):
            reads['r{}:{}'.format(i_s, i_r)] = mutate_structure(structure, mu_unpaired[i_s], mu_paired, len_seq)
            pop[i_s][i_r] = reads['r{}:{}'.format(i_s, i_r)]
    pop_avg = []
    for p,n_read in zip(pop,n_reads):
        mask = np.array(p >= SUB_A[0], dtype=np.uint8)
        pop_avg.append(mask.sum(axis=0)/n_read)
    return reads, pop_avg

def mutate_structure(structure, mu_unpaired, mu_paired, len_seq):
    """Mutate a structure.
    
    Inputs:
    -------
    
    structure: list of int.
        List of the unpaired bases.
    mu_unpaired: float
        Average mutation rate for unpaired bases.
    mu_paired: float
        Average mutation rate for paired bases.
    
    Output:
    -------
    
    mutated_read_bv: np.ndarray
        Array containing the mutated read under the byte format.
    """
    assert type(len_seq) 
    mutated_read_bv = np.ones(len_seq)
    for i in range(len_seq):
        if i in structure and (mutated_read_bv[i-3:i]==1).all():
            if np.random.random() < mu_unpaired:
                mutated_read_bv[i] = pick_random_mutation(not_insertion=i+1==len_seq)
        else:
            if np.random.random() < mu_paired:
                mutated_read_bv[i] = pick_random_mutation(not_insertion=i+1==len_seq)
    mutated_read_bv[mutated_read_bv==INS_3[0]+1] += INS_5[0]
    return mutated_read_bv

def pick_random_mutation(not_insertion):
    if not_insertion:
        return random.choices([DELET[0], SUB_A[0], SUB_C[0], SUB_G[0], SUB_T[0]], weights=(1, 10, 10, 10, 10))[0]
    return random.choices([DELET[0], INS_3[0], SUB_A[0], SUB_C[0], SUB_G[0], SUB_T[0]], weights=(1, 10, 10, 10, 10, 10))[0]


def create_real_structures(n_AC, n_struct, n_unpaired, n_shared, n_shared_3_structures):
    """Create a test dataset for the clustering algorithm.
    
    Inputs:
    -------
    
    n_AC: int
        Number of A and C bases in the sequence.
    n_struct: int
        Number of real alternative structures.
    n_unpaired: float
        Number of unpaired bases.
    n_shared: float
        Number of shared paired bases between the real alternative structures.
    n_shared_3_structures: string
        When n_struct = 3, the amount of bases shared between the three structures.
    
    Output:
    -------
    
    real_structures: list of list of int.
        List of the real alternative structures unpaired bases.
    """
    
    # create the sequence
    sequence = ["A"]*(n_AC//2) + ['C']*((n_AC+1)//2) + ['G']*((n_AC)//2) + ['T']*((n_AC+1)//2)
    np.random.shuffle(sequence)
    sequence = "".join(sequence)
    A_C_idx = set([i for i in range(len(sequence)) if sequence[i] in ["A", "C"]])
    
    # create a tool
    def pick(my_set, n):
        if len(my_set) < n:
            raise ValueError("Not enough bases available to create the real alternative structures. Check your ratios.")
        if n == 0:
            return set()
        return set(np.random.choice(list(my_set), n, replace=False))
    
    # create the real alternative structures
    real_structures = []
    
    if n_struct == 1:
        return [list(pick(A_C_idx, n_unpaired))], sequence
        
    elif n_struct == 2:
        common = pick(A_C_idx, n_shared)
        available_bases_idx = A_C_idx - common
        for i in range(n_struct):
            assert len(available_bases_idx) >= n_unpaired - n_shared, "Not enough bases available to create the real alternative structures. Check your ratios."
            real_structures.append(pick(available_bases_idx, n_unpaired - n_shared))
            real_structures[i] = set(real_structures[i]) | common
            available_bases_idx = set(available_bases_idx) - real_structures[i]
        return [list(real_structures[0]), list(real_structures[1])], sequence
    
    elif n_struct == 3:
        A = pick(A_C_idx, n_unpaired)
        B = pick(A_C_idx - A, n_unpaired - n_shared) | pick(A, n_shared)
        C = pick(A_C_idx - A - B, n_unpaired - n_shared*2 + n_shared_3_structures) 
        C = C | pick(A-B, n_shared - n_shared_3_structures)
        C = C | pick(B-A, n_shared - n_shared_3_structures)
        C = C | pick(A & B, n_shared_3_structures)
        
        return [list(A), list(B), list(C)], sequence

    else:
        raise ValueError("n_struct must be 1, 2 or 3.")
    

def generate_clustering(path_bv, path_json, n_AC, n_unpaired, n_shared, n_reads, mu_unpaired, mu_paired=0.01, n_shared_3_structures=0):
    """Create a test dataset for the clustering algorithm.
    
    Inputs:
    -------
    
    path_bv: string
        Path to the output file for the bitvector.
    path_json: string
        Path to the output file for the json.
    n_AC: int
        Number of A and C bases in the sequence. The sequence length will be twice this number.
    n_reads: list of int.
        Amount of reads per structure.
    n_unpaired: float
        Ratio of unpaired bases over the sequence length.
    n_shared: float
        Ratio of shared unpaired bases between the real alternative structures.
    mu_unpaired: list
        Average mutation rate for unpaired bases for each structure.
    mu_paired: float
        Average mutation rate for paired bases.
    n_shared_3_structures: string
        When len(n_reads) == 3, the amount of bases shared between the three structures.
    
    Output:
    -------
    
    bv: bitvectors under the form of an orc array.
    """
    
    n_struct = len(n_reads)
    assert n_struct in [1, 2, 3], "n_struct must be 1, 2 or 3."
    
    real_structures, sequence = create_real_structures(n_AC, n_struct, n_unpaired, n_shared, n_shared_3_structures)
    reads, pop_avg = create_reads_clustering(real_structures, mu_unpaired, mu_paired, n_reads, n_AC*2)
    
    df = pd.DataFrame.from_dict(reads, orient = 'index', columns=[c + str(i) for i, c in enumerate(sequence)], dtype=int)
    df['id'] = ['K' + str(c+1) + '_r' + str(s) for c in range(len(real_structures)) for s in range(n_reads[c])]

    sample_folder = '/'.join(path_bv.split('/')[:-1])
    if not os.path.exists(sample_folder):
        os.makedirs(sample_folder, exist_ok=True)

    df.to_orc(path_bv)
    print("Bitvector saved to", path_bv)
    return os.path.exists(path_bv)


def filter_range(series, ss, se):
    out = []
    for s in series:
        out.append([a for a in s if a >= ss and a < se])
    return out

def count_deletions(del_idx, sequence, ss, se, count_multiple_deletions = False):
    grouped_seq = group_sequence_by_base_idx(sequence[ss:se], offset=ss)
    count_del = np.zeros(se-ss, dtype=int)
    for g in grouped_seq:
        for d in del_idx:
            for i in g:
                if i in d:
                    if len(g) == 1:
                        count_del[i] += 1
                    elif count_multiple_deletions:
                        for j in g:
                            count_del[j] += 1
    return count_del.tolist()

def count_deletions_per_read(del_idx, sequence, ss, se, count_multiple_deletions = False):
    grouped_seq = group_sequence_by_base_idx(sequence[ss:se], offset=ss)
    out = np.zeros(len(del_idx), dtype=int)
    for i, d in enumerate(del_idx):
        for g in grouped_seq:
            if len(g) == 1:
                if g[0] in d:
                    out[i] += 1
            elif count_multiple_deletions:
                for j in g:
                    if j in d:
                        out[i] += 1
    return out

def count_insertions_per_read(ins_idx, sequence, ss, se):
    group_seq = group_sequence_by_base_idx(sequence[ss:se], offset=ss)
    out = np.zeros(len(ins_idx), dtype=int)
    for i, d in enumerate(ins_idx):
        for g in group_seq:
            if len(g) == 1:
                if g[0] in d:
                    out[i] += 1
            else:
                for dd in d:
                    if dd in g or dd+1 in g:
                        if next_base(sequence[dd]) == sequence[dd-1]:
                            g.append(dd-1)
                            for _ in g:
                                out[i] += 1
    return out

def count_substitution_per_read(sub_idx, ss, se):
    out = np.zeros(len(sub_idx), dtype=int)
    for i, d in enumerate(sub_idx):
        for j in d:
            if j >= ss and j < se:
                out[i] += 1
    return out

def generate_output_files(file, sample_profile, library, samples, clusters = None, rnastructure_config = None):
    if clusters is None:
        library = pd.read_csv(library)
        samples = pd.read_csv(samples)
        out = samples.to_dict('records')[0]
        for construct, v in sample_profile.items():
            out[construct] = {}
            out[construct]['num_reads'] = sample_profile[construct]['number_of_reads']
            out[construct]['num_aligned'] = sample_profile[construct]['number_of_reads']
            out[construct]['barcode'] = v['barcodes']
            out[construct]['barcode_start'] = v['barcode_start']
            out[construct]['some_random_attribute'] = library['some_random_attribute'].values[0]
            out[construct]['sequence'] = v['reference']
            for s, ss, se in zip(v['sections'], v['section_start'], v['section_end']):
                # DREEM
                insertions = filter_range(v['insertions'], ss, se)
                deletions = filter_range(v['deletions'], ss, se)
                mutations = filter_range(v['mutations'], ss, se)
                out[construct][s] = {}
                out[construct][s]['section_start'] = ss +1
                out[construct][s]['section_end'] = se
                out[construct][s]['sequence'] = v['reference'][ss:se]
                out[construct][s]['pop_avg'] = {}
                for base in ['A', 'C', 'G', 'T','N']:
                    out[construct][s]['pop_avg']['mod_bases_{}'.format(base)] = count_mut_mod(v['reference'], mutations, base, ss, se)
                out[construct][s]['pop_avg']['num_of_mutations'] = []#np.array( count_insertions_per_read(insertions, v['reference'], ss, se) + count_deletions_per_read(deletions, v['reference'], ss, se) + count_substitution_per_read(mutations, ss, se)).tolist()# +).tolist()# + count_mut_mod(v['reference'], mutations, 'N', ss, se)).sum()
                out[construct][s]['pop_avg']['del_bases'] =  count_deletions(deletions, v['reference'], ss, se)
                out[construct][s]['pop_avg']['ins_bases'] =  count_insertions(insertions,  v['reference'], ss, se)
                out[construct][s]['pop_avg']['mut_bases'] =  [int(m) for m in (np.array(out[construct][s]['pop_avg']['mod_bases_N']) + np.array(count_deletions(deletions, v['reference'], ss, se,count_multiple_deletions=1)) + np.array( out[construct][s]['pop_avg']['ins_bases']))]
                out[construct][s]['pop_avg']['cov_bases'] = count_coverage(v['number_of_reads'], v['insertions'], v['deletions'], ss, se, v['reference'])
                out[construct][s]['pop_avg']['info_bases'] = out[construct][s]['pop_avg']['cov_bases']
                out[construct][s]['pop_avg']['mut_rates'] = np.array( np.array(out[construct][s]['pop_avg']['mut_bases'])/np.array(out[construct][s]['pop_avg']['info_bases'])).tolist()
                out[construct][s]['pop_avg']['worst_cov_bases'] = min(out[construct][s]['pop_avg']['cov_bases'])
                out[construct][s]['pop_avg']['skips_short_reads'] = 0
                out[construct][s]['pop_avg']['skips_too_many_muts'] = 0
                out[construct][s]['pop_avg']['skips_low_mapq'] = 0

                # RNAstructure
                if rnastructure_config is not None:
                    rnastructure_config['temp_folder'] = os.path.join(os.path.dirname(file), 'temp')
                    os.makedirs(rnastructure_config['temp_folder'], exist_ok=True)
                    rna = RNAstructure(rnastructure_config)
                    mp = pd.Series({
                        'sequence': out[construct][s]['sequence'],
                        'construct': construct,
                        'section': s,
                        'info_bases': out[construct][s]['pop_avg']['info_bases'],
                        'mut_bases': out[construct][s]['pop_avg']['mut_bases'],
                        'temperature_k': out['temperature_k']
                    })
                    rna_pred = rna.run(mp, out['sample'], sequence_only = True)
                    for k, va in rna_pred.items():
                        out[construct][s][k] = va
                        
                    # Poisson
                    poisson_pred = poisson.compute_conf_interval(mp.info_bases, mp.mut_bases)
                    for k, va in poisson_pred.items():
                        out[construct][s]['pop_avg'][k] = va
                    os.system('rm -fr {}'.format(rnastructure_config['temp_folder']))

    else:       
        raise NotImplementedError('Clustering not implemented yet')

    out = sort_dict(out)
    options = jsbeautifier.default_options()
    options.indent_size = 4
    with open(file, 'w') as f:
        f.write(jsbeautifier.beautify(json.dumps(out), options))


def generate_files(sample_profile, module, inputs, outputs, test_files_dir, sample_name, rnastructure_config=None):
    """Generate the files for a sample profile.

    Parameters
    ----------
    sample_profile : dict
        The sample dictionary.
        Attributes are:
        - constructs
            - reads
            - number_of_reads
            - mutations
            - insertions
            - deletions
    module : str
        The module name.
    inputs : list
        The inputs files.
    outputs : list
        The outputs files.
    test_files_dir : str
        The test files directory.
    sample_name : str
        The sample name.
    rnastructure_config : str
        The RNAstructure config file.
    """
    # make input and output folders
    input_folder = os.path.join(test_files_dir, 'input', module, sample_name)
    output_folder = os.path.join(test_files_dir, 'expected_output', module, sample_name)
    if not os.path.exists(input_folder):
        os.makedirs(input_folder, exist_ok=True)
    if outputs != ['output']:
        if not os.path.exists(output_folder):
            os.makedirs(output_folder, exist_ok=True)
    else:
        os.makedirs(os.path.join(test_files_dir, 'expected_output', module), exist_ok=True)
    if ('bitvector' in inputs or 'bitvector' in outputs) and 'library' not in inputs:
        inputs.insert(0,'library')
    if 'library' in inputs:
        if inputs.index('library') > 0:
            inputs.insert(0, inputs.pop(inputs.index('library')))
    for inpt in inputs:
        generate_file_factory(input_folder, inpt, sample_profile)
    for output in outputs:
        generate_file_factory(input_folder, output, sample_profile,output_folder, rnastructure_config)

def generate_file_factory(input_path, file_type, sample_profile, output_path=None, rnastructure_config=None):
    if output_path is None:
        path = input_path
    else:
        path = output_path
    library = os.path.join(input_path, 'library.csv')
    samples = os.path.join(input_path, 'samples.csv')
    if file_type == 'library':
        generate_library_file(library, sample_profile)
    elif file_type == 'fasta':
        generate_fasta_file(os.path.join(path, 'reference.fasta'), sample_profile)
    elif file_type == 'fastq':
        generate_fastq_files(path, sample_profile)
    elif file_type == 'bitvector':
        generate_bitvector_files(path, sample_profile, library)
    elif file_type == 'samples':
        generate_samples_file(samples)
    elif file_type == 'demultiplexed_fastq':
        generate_demultiplexed_fastq_files(path, sample_profile)
    elif file_type == 'sam':
        generate_sam_files(path, sample_profile)
    elif file_type == 'bam':
        generate_sam_files(path, sample_profile)
        generate_bam_files(path, sample_profile)
    elif file_type == 'output':
        generate_output_files(path+'.json', sample_profile, library, samples, None, rnastructure_config) # TODO: add clustering
    elif file_type == 'clustering':
        for params in sample_profile.values():
            generate_clustering(**params)
    else:
        raise ValueError('File type not recognized: "{}"'.format(file_type))

def assert_files_exist(sample_profile, module, files_types, in_out_pred_dir, sample_name):
    """Assert that the files for a sample profile exist.

    Parameters
    ----------
    sample_profile : dict
        The sample dictionary.
        Attributes are:
        - constructs
            - reads
            - number_of_reads
            - mutations
            - insertions
            - deletions
    module : str
        The module name.
    files_types : list
        The files types. Ex: bitvector, fasta, fastq, ...
    in_out_pred_dir : str
        The input, output or expected output directory.
    sample_name : str
        The sample name.
    """
    # make input and output folders
    folder = os.path.join(in_out_pred_dir, module, sample_name)
    if not os.path.isfile(folder) and 'output' not in files_types:
        assert os.path.exists(folder), 'Folder of {} files does not exist: {}'.format(folder.split('/')[-3], folder)
    
    for file in files_types:
        assert_file_factory(folder, file, sample_profile, sample_name)
    
        
def assert_file_factory(path, file_type, sample_profile, sample_name):
    if file_type == 'library':
        assert os.path.exists(os.path.join(path, 'library.csv')), 'Library file does not exist: {}'.format(os.path.join(path, 'library.csv'))
    elif file_type == 'fasta':
        assert os.path.exists(os.path.join(path, 'reference.fasta')), 'Fasta file does not exist: {}'.format(os.path.join(path, 'reference.fasta'))
    elif file_type == 'fastq':
            for read in [1,2]:
                assert os.path.exists(os.path.join(path, '{}_R{}.fastq'.format(sample_name, read))), 'Fastq file does not exist: {}'.format(os.path.join(path, '{}_R{}.fastq'.format(sample_name, read)))
    elif file_type == 'bitvector':
        for construct in sample_profile:
            for section in sample_profile[construct]['sections']:
                assert os.path.exists(os.path.join(path, '{}/{}/0.orc'.format(construct, section))), 'Bitvector file does not exist: {}'.format(os.path.join(path, '{}/{}/0.orc'.format(construct, section)))
    elif file_type == 'clustering':
        for file in sample_profile:
            assert os.path.isfile(os.path.join(path, '{}/0.orc'.format(file))), 'Bitvector file does not exist: {}'.format(os.path.join(path, '{}/0.orc'.format(file)))
    elif file_type == 'samples':
        assert os.path.isfile(os.path.join(path, 'samples.csv')), 'Samples csv file does not exist: {}'.format(os.path.join(path, 'samples.csv'))
    elif file_type == 'demultiplexed_fastq':
        for construct in sample_profile:
            for read in [1,2]:
                assert os.path.exists(os.path.join(path, '{}_R{}.fastq'.format(construct, read))), 'Demultiplexed fastq file does not exist: {}'.format(os.path.join(path, '{}_R{}.fastq'.format(construct, read)))
    elif file_type == 'sam':
        for construct in sample_profile:
            assert os.path.exists(os.path.join(path,'{}.sam'.format(construct))), 'Sam file does not exist: {}'.format(os.path.join(path, '{}.sam'.format(construct)))
    elif file_type == 'bam':
        for construct in sample_profile:
            assert os.path.exists(os.path.join(path,'{}.bam'.format(construct))), 'BAM file does not exist: {}'.format(os.path.join(path, '{}.bam'.format(construct)))


from dreem.draw import manipulator, util, plotter
import pandas as pd
import numpy as np
from dreem.util.dump import sort_dict, flatten_json

class Study(object):
    """A class to store information about a study, i.e a set of samples that are relevant to be studied together.

    Attributes:
        name (str, optional): Short description (<~20 char) of your study. Defaults to None.
        samples (List[str], optional): Names of your study's samples. Defaults to None.
        
    Example:
        >>> study = Study('example',['A1', 'B2', 'B3'])
    """

    attr_list = ['name','samples']

    def __init__(self, data=None, samples=None, min_cov_bases=0, filter_by='sample') -> None:
        """Creates a Study object.

        Args:
            data (dict or list[dict], optional): A dictionary or list of dictionaries containing the data to be loaded. One dictionary corresponds to a fastq file. Defaults to None.
            samples (List[str], optional): List of samples to load. Defaults to None.
            min_cov_bases (int, optional): Minimum number of base coverage for a row to be filtered-in. Defaults to 0.
            filter_by (str, optional): Filter rows by sample or study. When filtered by study, if a row passes the filter, rows with the same 'construct', 'section' and 'cluster' fields for all other samples have a sufficient base coverage. Defaults to 'sample'.            

        Example:
            >>> study = Study(data = {'sample':'mysample',{'construct1': {'section1': {'cluster1': {'mut_bases': [100], 'cov_bases': [1000]}}}}},
                              samples=['A1', 'B2', 'B3'], 
                              min_cov_bases=1000, 
                              filter_by='study')
        """
        self.samples = samples
        if data is not None:
            if self.samples is None:
                self.samples = [d['sample'] for d in data]
            self.df = pd.DataFrame()
            if type(data) is not list:
                data = [data]
            print('Turning data into a dataframe...')
            for sample in data:
                print(sample['sample'], end='... ')
                self.df = pd.concat([self.df, pd.DataFrame(flatten_json(sort_dict(sample)))], axis=0)
            print('Done.')
            print('Setting dataframe...')
            self.set_df(self.df, min_cov_bases=min_cov_bases, filter_by=filter_by, samples=samples)
            print('Done.')
        else:
            self.df = None

    @classmethod
    def from_dict(cls, di:dict):
        """Set attributes of this Study object from a dictionary.

        Args:
            di (dict): a dictionary containing keys such as ['name','description','samples','label','conditions'].

        Returns:
            Study: a study object.

        Example:
        >>> di = {'name':'temperature','samples':['A1','B2','B3']}
        >>> study = Study().from_dict(di)
        >>> print(study.name, study.samples)
        temperature ['A1', 'B2', 'B3']
        """
        for attr in cls.attr_list:
            try: 
                di[attr]
            except: 
                di[attr]=None
        return cls(di['name'], di['samples'])

    def set_df(self, df, min_cov_bases=0, filter_by='sample', samples=None):
        df.reset_index(inplace=True, drop=True)
        self.df = df
        
        if not 'worst_cov_bases' in self.df.columns:
            self.df['worst_cov_bases'] = self.df['cov_bases'].apply(lambda x: min(x))
        
        self.df = self.df[self.df['worst_cov_bases'] >= min_cov_bases]
        if samples is not None:
            assert type(samples) is list, 'samples must be a list'
            self.df = self.df[self.df['sample'].isin(samples)]
        
        if filter_by == 'study':
            self.filter_by_study(inplace=True)
        
        for attr in ['sample','construct']:
            self.df[attr] = self.df[attr].astype(str)
        
        for attr in ['section','cluster']:
            if attr not in self.df.columns:
                self.df[attr] = 0
        
        self.df['deltaG'] = self.df['deltaG'].apply(lambda x: 0.0 if x == 'void' else float(x))
    
    
    def filter_by_study(self, inplace=False):
        df = self.df.groupby(['construct', 'section', 'cluster']).filter(lambda x: len(self.df['sample'].unique()) == len(x['sample'].unique()))
        if inplace:
            self.df = df
        return df.copy()

    def get_df(self, **kwargs):
        return manipulator.get_df(self.df, **kwargs)

    def get_samples(self):
        return self.df['sample'].unique()

    def get_constructs(self, sample:str):
        return self.df[self.df['sample'] == sample]['construct'].unique()

    def get_sections(self, sample:str, construct:str):
        return self.df[(self.df['sample'] == sample) & (self.df['construct'] == construct)]['section'].unique()

    def get_clusters(self, sample:str, construct:str, section:str):
        return self.df[(self.df['sample'] == sample) & (self.df['construct'] == construct)& (self.df['section'] == section)]['cluster'].unique()
       
    def load_studies(studies_file_path:str):
        return load_studies(studies_file_path)



    def mutation_fraction(self, **kwargs)->dict:
        """Plot the mutation rates as histograms.

        Args:
            sample (list, int, str, optional): Filter rows by sample (list of samples or just a sample). Defaults to None.
            construct (list, int, str, optional): Filter rows by construct (list of constructs or just a construct). Defaults to None.
            section (list, int, str, optional): Filter rows by section (list of sections or just a section). Defaults to None.
            cluster (list, int, str, optional): Filter rows by cluster (list of clusters or just a cluster). Defaults to None.
            base_index (list, int, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base index. Can be a unique sequence in the row's sequence, a list of indexes or a single index. Defaults to None.
            base_type (list, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base type. Defaults to ['A','C','G','T'].
            base_pairing (bool, optional): Filter per-base attributes (mut_rates, sequence, etc) by expected base pairing. See RNAstructure_use_XXX arguments. Defaults to None.
            RNAstructure_use_DMS (bool, optional): Use DMS for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            RNAstructure_use_temp (bool, optional): Use temperature for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            show_ci(bool, optional): Show confidence intervals. Defaults to True.
            savefile(str, optional): Path to save the plot. Defaults to None.
            use_iplot(bool, optional): Use iplot instead of plot (for Jupyter notebooks). Defaults to True.
            title(str, optional): Title of the plot. Defaults to None, in which case a standard name is given.

        Returns:
            dict: Figure and data of the output plot.
        """

        return plotter.mutation_fraction(manipulator.get_df(self.df, **{k:v for k,v in kwargs.items() if k in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)}), **{k:v for k,v in kwargs.items() if k in plotter.mutation_fraction.__code__.co_varnames})

    def deltaG_vs_mut_rates(self, **kwargs)->dict:
        """Plot the mutation rate of each paired-expected base of the ROI for each construct of a sample, w.r.t the deltaG estimation.

        Args:
            sample (list, int, str, optional): Filter rows by sample (list of samples or just a sample). Defaults to None.
            construct (list, int, str, optional): Filter rows by construct (list of constructs or just a construct). Defaults to None.
            section (list, int, str, optional): Filter rows by section (list of sections or just a section). Defaults to None.
            cluster (list, int, str, optional): Filter rows by cluster (list of clusters or just a cluster). Defaults to None.
            min_cov_bases (int, optional): Filter rows by a minimum threshold for base coverage. Defaults to 0.
            base_index (list, int, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base index. Can be a unique sequence in the row's sequence, a list of indexes or a single index. Defaults to None.
            base_type (list, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base type. Defaults to ['A','C','G','T'].
            base_pairing (bool, optional): Filter per-base attributes (mut_rates, sequence, etc) by expected base pairing. See RNAstructure_use_XXX arguments. Defaults to None.
            RNAstructure_use_DMS (bool, optional): Use DMS for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            RNAstructure_use_temp (bool, optional): Use temperature for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            savefile(str, optional): Path to save the plot. Defaults to None.
            use_iplot(bool, optional): Use iplot instead of plot (for Jupyter notebooks). Defaults to True.
            title(str, optional): Title of the plot. Defaults to None, in which case a standard name is given.
            models (List[str], optional): Models to fit on the data using scipy.optimize.curve_fit. Under the form ``'lambda x, a, b: a*x+b'`` where ``x`` is the variable. Defaults to [].
            **kwargs: Additional arguments to pass to filter rows by. Ex: flank='flank_1' will keep only rows with flank=flank_1. 

        Returns:
            dict: Figure and data of the output plot.
        """
        return plotter.deltaG_vs_mut_rates(manipulator.get_df(self.df, **{k:v for k,v in kwargs.items() if k in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)}), **{k:v for k,v in kwargs.items() if k in plotter.deltaG_vs_mut_rates.__code__.co_varnames})

    
    def exp_variable_across_samples(self, **kwargs)->dict:
        """Plot the mutation rate of each paired-expected base of the ROI for each construct of a sample, w.r.t the deltaG estimation.

        Args:
            construct (str): Construct of your row.
            experimental_variable (str): x axis column value, must be a per-sample attribute.
            section (list, int, str, optional): Filter rows by section (list of sections or just a section). Defaults to None.
            cluster (str): Cluster of your row.
            structure (str, optional): Structure to use for base_paired filtering. Defaults to 'structure'.
            index (_type_, optional): Indexes to plot. Defaults to ``'all'``.
            base_type (List[str], optional): Bases type to plot. Defaults to ``['A','C','G','T']``.
            max_mutation (float, optional): Maximum mutation rate to plot. Defaults to 0.15.
            RNAstructure_use_DMS (bool, optional): Use DMS for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            RNAstructure_use_temp (bool, optional): Use temperature for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            models (List[str], optional): Models to fit on the data using scipy.optimize.curve_fit. Under the form ``'lambda x, a, b: a*x+b'`` where ``x`` is the variable. Defaults to [].
            savefile (str, optional): Path to save the plot. Defaults to None.
            use_iplot(bool, optional): Use iplot instead of plot (for Jupyter notebooks). Defaults to True.
            title(str, optional): Title of the plot. Defaults to None, in which case a standard name is given.

        Returns:
            dict: Figure and data of the output plot.
        """
        return plotter.exp_variable_across_samples(manipulator.get_df(self.df, **{k:v for k,v in kwargs.items() if k in  list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)}), **{k:v for k,v in kwargs.items() if k in plotter.exp_variable_across_samples.__code__.co_varnames})
    
    def auc(self, **kwargs)->dict:
        """Plot the AUC for each mutation profile of the selected data. 

        Args:
            sample (list, int, str, optional): Filter rows by sample (list of samples or just a sample). Defaults to None.
            construct (list, int, str, optional): Filter rows by construct (list of constructs or just a construct). Defaults to None.
            section (list, int, str, optional): Filter rows by section (list of sections or just a section). Defaults to None.
            cluster (list, int, str, optional): Filter rows by cluster (list of clusters or just a cluster). Defaults to None.
            min_cov_bases (int, optional): Filter rows by a minimum threshold for base coverage. Defaults to 0.
            base_index (list, int, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base index. Can be a unique sequence in the row's sequence, a list of indexes or a single index. Defaults to None.
            base_type (list, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base type. Defaults to ['A','C','G','T'].
            base_pairing (bool, optional): Filter per-base attributes (mut_rates, sequence, etc) by expected base pairing. See RNAstructure_use_XXX arguments. Defaults to None.
            RNAstructure_use_DMS (bool, optional): Use DMS for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            RNAstructure_use_temp (bool, optional): Use temperature for the RNAstructure prediction when filtering by base pairing and predicting deltaG. Defaults to False.
            savefile(str, optional): Path to save the plot. Defaults to None.
            use_iplot(bool, optional): Use iplot instead of plot (for Jupyter notebooks). Defaults to True.
            title(str, optional): Title of the plot. Defaults to None, in which case a standard name is given.
            **kwargs: Additional arguments to pass to filter rows by. Ex: flank='flank_1' will keep only rows with flank=flank_1. 
    
        Returns:
            dict: Figure and data of the output plot.
        """
        return plotter.auc(manipulator.get_df(self.df, **{k:v for k,v in kwargs.items() if k in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)}), **{k:v for k,v in kwargs.items() if k in plotter.auc.__code__.co_varnames})


    def mutation_fraction_delta(self, **kwargs)->dict:
        """Plot the mutation rate difference between two mutation profiles.
        sample1: sample of the first mutation profile.
        sample2: sample of the second mutation profile.
        construct1: construct of the first mutation profile.
        construct2: construct of the second mutation profile.
        section1: section of the first mutation profile.
        section2: section of the second mutation profile.
        cluster1: cluster of the first mutation profile.
        cluster2: cluster of the second mutation profile.
        base_index1: base index of the first mutation profile.
        base_index2: base index of the second mutation profile.
        base_type1: base type of the first mutation profile.
        base_type2: base type of the second mutation profile.
        base_pairing1: base pairing of the first mutation profile.
        base_pairing2: base pairing of the second mutation profile.
        savefile(str, optional): Path to save the plot. Defaults to None.
        use_iplot(bool, optional): Use iplot instead of plot (for Jupyter notebooks). Defaults to True.
        title(str, optional): Title of the plot. Defaults to None, in which case a standard name is given.        
        """
        
        df1 = manipulator.get_df(self.df, **{k[:-1]:v for k,v in kwargs.items() if k.endswith('1') and k[:-1] in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)})
        assert len(df1)>0, 'No rows found for the first mutation profile.'
        assert len(df1)==1, 'More than one row found for the first mutation profile.'
        df2 = manipulator.get_df(self.df, **{k[:-1]:v for k,v in kwargs.items() if k.endswith('2') and k[:-1] in list(self.df.columns)+ list(manipulator.get_df.__code__.co_varnames)})
        assert len(df2)>0, 'No rows found for the second mutation profile.'
        assert len(df2)==1, 'Only one row should be selected for the second mutation profile.'
        return plotter.mutation_fraction_delta(pd.concat([df1, df2]).reset_index(drop=True), **{k:v for k,v in kwargs.items() if k in plotter.mutation_fraction_delta.__code__.co_varnames})

    def base_coverage(self, **kwargs):
        """Plot the base coverage of several constructs in a sample.

        Args:
            samp (str): Sample of your rows.
            constructs (List[str]): Constructs of your rows.
            section (str): Region of your row.
            cluster (int, optional): Cluster of your row. Defaults to 0. 
            index (_type_, optional): Indexes to plot. Defaults to ``'all'``. Can be a series of 0-indexes (ex: [43,44,45,48]), 'roi', 'all', or a unique sequence (ex: 'ATTAC')
            base_type (List[str], optional): Bases type to plot. Defaults to ``['A','C','G','T']``.
            base_paired (bool, optional): Base-pairing predicition to plot. Defaults to None.
            structure (str, optional): Structure to use for base_paired filtering. Defaults to 'structure'.
            show_ci (bool, optional): Show confidence interval on the histogram. Defaults to True.
            savefile (str, optional): Path to save the plot. Defaults to None.
            use_iplot(bool, optional): Use iplot instead of plot (for Jupyter notebooks). Defaults to True.
            title(str, optional): Title of the plot. Defaults to None, in which case a standard name is given.

        Returns:
            dict: Figure and data of the output plot.

        """
        return 0# plotter.base_coverage(self._df, **kwargs)



def load_studies(studies_file_path:str)->dict[str:Study]:
    """Read formatted file with samples, and turn it into a dataframe containing studies.

    Args:
        studies_file_path (str): path+title of the csv file containing the samples.

    Returns:
        (pd.DataFrame): studies of the csv file, indexed by study.
    """

    studies_dict, studies_data = {}, pd.read_csv(studies_file_path)

    for col in studies_data.groupby('name')[Study.attr_list]:
        solo_item = lambda x: x[0] if len(set(x)) == 1 else x  
        studies_dict[col[0]] = {attr: solo_item(list(col[1][attr])) for attr in (Study.attr_list)} 

    return {k:Study.from_dict(v) for k,v in studies_dict.items()}



import click
from main import run
from click_option_group import optgroup
from dreem.util.cli import *

@click.command()

@optgroup.group('I/O')
@input_dir
@opti_library
@opto_out_dir
@samples
@sample
@clustering_file

@optgroup.group('Selection')
@coords
@primers
@fill

@optgroup.group('RNAstructure')
@rnastructure_path
@rnastructure_temperature
@rnastructure_fold_args
@rnastructure_dms
@rnastructure_dms_min_unpaired_value
@rnastructure_dms_max_paired_value
@rnastructure_partition
@rnastructure_probability

@optgroup.group('Miscellaneous')
@poisson
@verbose


def cli(**args):
    run(**args)

if __name__ == '__main__':
    run()



import json
import os
import random
import shlex
import subprocess
import sys
from tempfile import NamedTemporaryFile
from typing import List, Set, Dict
import yaml

import numpy as np
import pandas as pd
import pyarrow.orc  # This prevents: AttributeError: module 'pyarrow' has no attribute 'orc'

from dreem.aggregate import poisson


# CONSTANTS
BASES = b"ACGT"
COMPS = b"TGCA"
RBASE = b"ACGU"
RCOMP = b"UGCA"
BASEN = b"N"
BASES_SET = set(BASES)
BASEN_SET = set(BASES + BASEN)

# BYTE ENCODINGS FOR MUTATION VECTORS
BLANK = b"\x00"  # 00000000 (000): no coverage at this position
MATCH = b"\x01"  # 00000001 (001): match with reference
DELET = b"\x02"  # 00000010 (002): deletion from reference
INS_5 = b"\x04"  # 00000100 (004): insertion 5' of base in reference
INS_3 = b"\x08"  # 00001000 (008): insertion 3' of base in reference
SUB_A = b"\x10"  # 00010000 (016): substitution to A
SUB_C = b"\x20"  # 00100000 (032): substitution to C
SUB_G = b"\x40"  # 01000000 (064): substitution to G
SUB_T = b"\x80"  # 10000000 (128): substitution to T
AMBIG = b"\xff"  # 11111111 (255): could be anything
SUB_N = (SUB_A[0] | SUB_C[0] | SUB_G[0] | SUB_T[0]).to_bytes()
ANY_N = (MATCH[0] | SUB_N[0]).to_bytes()
MADEL = (MATCH[0] | DELET[0]).to_bytes()
NOSUB = (AMBIG[0] ^ SUB_N[0]).to_bytes()
UNAMB_SET = set(b"".join([BLANK, MATCH, DELET, INS_5, INS_3,
                          SUB_A, SUB_C, SUB_G, SUB_T]))

BLANK_INT = 0
MATCH_INT = 1
DELET_INT = 2
INS_5_INT = 4
INS_3_INT = 8
SUB_A_INT = 16
SUB_C_INT = 32
SUB_G_INT = 64
SUB_T_INT = 128
SUB_N_INT = SUB_A_INT + SUB_C_INT + SUB_G_INT + SUB_T_INT
AMBIG_INT = 255

B_MATCH = 0
B_DELET = 1
B_INS_5 = 2
B_INS_3 = 3
B_SUB_A = 4
B_SUB_C = 5
B_SUB_G = 6
B_SUB_T = 7

# COMMANDS
BOWTIE2_CMD = "bowtie2"
BOWTIE2_BUILD_CMD = "bowtie2-build"
CUTADAPT_CMD = "cutadapt"
FASTQC_CMD = "fastqc"
PASTE_CMD = "paste"
SAMTOOLS_CMD = "samtools"

# GLOBAL SETTINGS
DEFAULT_PROCESSES = cpus if (cpus := os.cpu_count()) else 1
BASE_COLORS = {"A": "#D3822A", "C": "#5AB4E5", "G": "#EAE958", "T": "#357766"}
DEFAULT_PHRED_ENCODING = 33
OUTPUT_DIR = "output"
TEMP_DIR = "temp"


def make_folder(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)
    return folder

def make_cmd(args, module):
    cmd = 'dreem-' + module + ' '
    for key, value in args.items():
        if type(value) in (list, tuple):
            for v in value:
                cmd += '--' + key + ' ' + str(v) + ' '
        elif value is not None:
            cmd += '--' + key + ' ' + str(value) + ' '
    return cmd

def run_cmd(args: List[str], shell: bool = False):
    os.system(' '.join(args))#, check=True, shell=shell)
    cmd = shlex.join(args)
    return cmd

def name_temp_file(dirname: str, prefix: str, suffix: str):
    file = NamedTemporaryFile(dir=dirname, prefix=prefix, suffix=suffix,
                              mode="w", delete=False)
    file.close()
    return file.name

def get_filename(file_path: str):
    return os.path.splitext(os.path.basename(file_path))[0]

def switch_directory(old_path: str, new_dir: str):
    return os.path.join(new_dir, os.path.basename(old_path))

def try_remove(file: str):
    try:
        os.remove(file)
    except OSError:
        pass

def try_rmdir(dir: str):
    try:
        os.rmdir(dir)
    except OSError:
        pass

def get_files(folder: str, ext: str):
    paths = []
    for construct in os.listdir(folder):
        if os.path.isdir(os.path.join(folder, construct)):
            for section in os.listdir(os.path.join(folder, construct)):
                if not section.endswith(ext):
                    continue
                paths.append(os.path.join(folder, construct, section))
    else:
        if construct.endswith(ext):
            paths.append(os.path.join(folder, construct))
    return paths


class Seq(bytes):
    __slots__ = []

    alph = b""
    comp = b""
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def __init__(self, seq: bytes):
        self.validate_seq(seq)
        super().__init__()
    
    @classmethod
    def validate_seq(cls, seq):
        if not seq:
            raise ValueError("seq is empty")
        if set(seq) - cls.alphaset:
            raise ValueError(f"Invalid characters in seq: '{seq.decode()}'")

    @property
    def rc(self):
        return self.__class__(self[::-1].translate(self.trans))

    def __getitem__(self, item):
        return self.__class__(super().__getitem__(item))

    def __str__(self):
        return self.decode()


class DNA(Seq):
    alph = BASES
    comp = COMPS
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def tr(self):
        """
        Transcribe DNA into RNA.
        """
        return RNA(self.replace(b"T", b"U"))


class RNA(Seq):
    alph = RBASE
    comp = RCOMP
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def rt(self):
        """
        Reverse transcribe RNA into DNA.
        """
        return DNA(self.replace(b"U", b"T"))


class FastaIO(object):
    __slots__ = ["_path", "_refs"]

    defsymbol = b">"
    deftrunc = len(defsymbol)

    def __init__(self, path: str):
        self._path = path



class FastaParser(FastaIO):
    def __init__(self, path: str):
        super().__init__(path)
        self._refs: Set[bytes] = set()

    @classmethod
    def _parse_fasta_record(cls, fasta, line: bytes):
        if not line.startswith(cls.defsymbol):
            raise ValueError("FASTA definition line does not start with "
                             f"'{cls.defsymbol.decode()}'")
        name = line.split()[0][cls.deftrunc:]
        seq = bytearray()
        while (line := fasta.readline()) and not line.startswith(cls.defsymbol):
            seq.extend(line.rstrip())
        seq = DNA(bytes(seq))
        return line, name, seq

    def parse(self):
        with open(self._path, "rb") as f:
            line = f.readline()
            while line:
                line, name, seq = self._parse_fasta_record(f, line)
                if name in self._refs:
                    raise ValueError(
                        f"Duplicate entry in {self._path}: '{name.decode()}'")
                self._refs.add(name)
                yield name, seq


class FastaWriter(FastaIO):
    def __init__(self, path: str, refs: Dict[bytes, DNA]):
        super().__init__(path)
        self._refs = refs
    
    def write(self):
        with open(self._path, "wb") as f:
            for ref, seq in self._refs.items():
                f.write(b"".join(self.defsymbol, ref, b"\n", seq, b"\n"))


def fastq_to_df(fastq_file):    
    df, data = pd.DataFrame(), pd.read_csv(fastq_file, sep='\t', header=None)
    df['construct'] = data.iloc[np.arange(0,len(data),4)].reset_index(drop=True)
    df['sequence'] = data.iloc[np.arange(1,len(data),4)].reset_index(drop=True)
    df['quality'] = data.iloc[np.arange(3,len(data),4)].reset_index(drop=True)
    return df


def sam_to_df(path):
    skiprows = 0
    with open(path) as f:
        while f.readline().startswith('@'):
            skiprows += 1
    df = pd.read_csv(path, sep='\t', skiprows=skiprows, header=None)
    df = df[df.columns[:11]]
    df.columns = ['QNAME', 'FLAG', 'RNAME', 'POS', 'MAPQ', 'CIGAR', 'RNEXT', 'PNEXT','TLEN', 'SEQ', 'QUAL']
    return df


def sort_fastq_pairs(fq1s, fq2s):
    if type(fq1s) is str:
        fq1s = [fq1s]
    if type(fq2s) is str:
        fq2s = [fq2s]
    assert len(fq1s) >= len(fq2s), 'More fq2s than fq1s'
    for f2 in fq2s:
        if f2.replace('_R2', '_R1') not in fq1s:
            raise ValueError(f'No matching pair for {f2}')
    fq1s = [f2.replace('_R2', '_R1') for f2 in fq2s] + [f for f in fq1s if f.replace('_R1','_R2') not in fq2s]
    fq2s += [None for f in fq1s if f.replace('_R1','_R2') not in fq2s]
    samples = [f.split('/')[-1].split('.')[0].split('_')[:-1] for f in fq1s]
    samples = ['_'.join(s) for s in samples]
    return fq1s, fq2s, samples


def query_muts(muts: np.ndarray, bits: int, sum_up = True, axis=0, set_type = 'superset'):
    """
    Count the number of times a query mutation occurs in each column
    or one column of a set of mutation vectors.
    The counting operation comprises three steps:
    1. bitwise AND to confirm at least one "1" bit is shared, e.g.
       bits: 11110000 & muts: 00100000 -> 00100000 (True)
       bits: 11110000 & muts: 00100010 -> 00100000 (True)
       bits: 11110000 & muts: 00000000 -> 00000000 (False)
    2. bitwise OR to confirm no "1" bit in muts is not in bits, e.g.
       bits: 11110000 | muts: 00100000 -> 11110000 =? 11110000 (True)
       bits: 11110000 | muts: 00100010 -> 11110010 =? 11110000 (False)
       bits: 11110000 | muts: 00000000 -> 11110000 =? 11110000 (True)
    3. logical AND to confirm that both tests pass, e.g.
       bits: 11110000, muts: 00100000 -> True  AND True  (True)
       bits: 11110000, muts: 00100010 -> True  AND False (False)
       bits: 11110000, muts: 00000000 -> False AND True  (False)

    Arguments
    muts: NDArray of a set of mutation vectors (2-dimensional)
          or one column in a set of mutation vectors (1-dimensional).
          Data type must be uint8.
    bits: One-byte int in the range [0, 256) representing the mutation
          to be queried. The bits in the int encode the mutation as
          defined above, e.g.
          - 00000010 (int 2) is a deletion
          - 11010001 (int 209) is either substitution to A, G, or T
                               or a match to C
    
    Returns
    if sum_up: 
        int of the number of times the query mutation occurs in muts
        count: If muts is 1-dimensional, int of the number of times the
            query mutation occurs in muts.
            If muts is 2-dimensional, NDArray with one int for each
            column in muts.
    if not sum_up:
        bool NDArray with one bool for each column in muts.
        True if the query mutation occurs in the column, False otherwise.
    """
    if not muts.dtype == np.uint8:
        raise TypeError('muts must be of type uint8 and not {}'.format(muts.dtype))
    #print(np.logical_and(muts & bits, (muts | bits) == bits).sum(axis=0))
    assert isinstance(bits, int) and 0 <= bits < 256

        
    if set_type == 'subset':
        logic_fun = lambda m, b: np.logical_and(m & b, (m | b) == b)
    if set_type == 'superset':
        logic_fun = lambda m, b: np.array(m & b, dtype=bool)
    
    if sum_up:
        return logic_fun(muts, bits).sum(axis=axis)
    else:
        return logic_fun(muts, bits)
     




import os
from dreem import util
import numpy as np
import pandas as pd
from tqdm import tqdm
import subprocess
from dreem.util.cmd import run_cmd
                
class RNAstructure(object): 
    #TODO adapt the config to the new config which is the class attributes
    def __init__(self, config) -> None:
        """
        config: dict
            path: str
                Path to RNAstructure, to predict structure and free energy.
            temperature: bool
                Use sample.csv temperature values for RNAstructure.
            fold_args: str
                Arguments to pass to RNAstructure fold.
            dms: bool
                Use the DMS signal to amke predictions with RNAstructure.
            dms_min_unpaired_value: int
                Minimum unpaired value for using the dms signal as an input for RNAstructure.
            dms_max_paired_value: int
                Maximum paired value for using the dms signal as an input for RNAstructure.
            partition: bool
                Use RNAstructure partition function to predict free energy.
            probability: bool
                Use RNAstructure probability to predict free energy.
        """
        self.config = config
        self.rnastructure_path = os.path.abspath(config['path'])+'/'
        self.temperature = config['temperature']
        self.fold_args = config['fold_args']
        self.dms = config['dms']
        self.dms_min_unpaired_value = config['dms_min_unpaired_value']
        self.dms_max_paired_value = config['dms_max_paired_value']
        self.partition = config['partition']
        self.probability = config['probability']

    def make_files(self, temp_prefix):
        self.pfs_file = f"{temp_prefix}.pfs"
        self.ct_file = f"{temp_prefix}.ct"
        self.dot_file = f"{temp_prefix}_dot.txt"
        self.fasta_file = temp_prefix+'.fasta'
        self.prob_file = temp_prefix+'_prob.txt'

    def create_fasta_file(self, construct, sequence):
        # push the ref into a temp file
        temp_fasta = open(self.fasta_file, 'w')
        temp_fasta.write('>'+construct+'\n'+sequence)
        temp_fasta.close()

    def predict_construct(self, use_dms=False, dms_file=False, use_temperature=False, temperature_k=None):
        # Run RNAstructure
        suffix = ''
        if use_temperature:
            suffix += f' --temperature {temperature_k}'
        if use_dms:
            suffix = f' -dms {dms_file}'
        cmd = f"{self.rnastructure_path}Fold {self.fasta_file} {self.ct_file}" + suffix
        run_cmd(cmd.split(' '))
        assert os.path.getsize(self.ct_file) != 0, f"{self.ct_file} is empty, check that RNAstructure works"

    # cast the temp file into a dot_bracket structure and extract the attributes
    def extract_deltaG_struct(self):
        run_cmd(f"ct2dot {self.ct_file} 1 {self.dot_file}".split(' '))
        temp_dot = open(self.dot_file, 'r')
        first_line = temp_dot.readline().split()
        # If only dots in the structure, no deltaG 
        if len(first_line) == 4:
            _, _, deltaG, _ = first_line
            deltaG = float(deltaG)
        if len(first_line) == 1:
            deltaG, _ = 'void', first_line[0][1:]

        sequence = temp_dot.readline()[:-1] #  Remove the \n
        structure = temp_dot.readline()[:-1] # Remove the \n
        return deltaG, structure

    def generate_normalized_mut_rates(self,temp_prefix, info_bases, mut_bases):
        mut_rates = np.array(mut_bases)/np.array(info_bases) 
        mut_rates = [max(r,self.config['dms_max_paired_value']) - self.config['dms_max_paired_value'] for r in mut_rates]     
        mut_rates = np.array([min(r,self.config['dms_min_unpaired_value']) for r in mut_rates])  
        pd.DataFrame((mut_rates)/(max(mut_rates)-min(mut_rates)),index=list(range(1,1+len(mut_rates))))\
                    .to_csv(temp_prefix+'_DMS_signal.txt', header=False, sep='\t')

    def predict_ensemble_energy(self):
        cmd = f"{self.rnastructure_path}EnsembleEnergy {self.fasta_file} --DNA --sequence"
        splitted_output =subprocess.check_output(cmd.split(' ')).decode('utf-8').split(' ')
        return float(splitted_output[splitted_output.index(f"kcal/mol\n\nEnsemble")-1])

    def predict_mut_probability(self, use_temperature, temperature_k):

        cmd = f"{self.rnastructure_path}partition {self.fasta_file} {self.pfs_file} --DNA"
        if use_temperature:
            cmd += ' --temperature '+str(temperature_k)
        run_cmd(cmd.split(' '))
        run_cmd([self.rnastructure_path+'ProbabilityPlot', self.pfs_file,'-t',self.prob_file])
        with open(self.prob_file,"r") as f:
            lines=f.readlines()
            out={'i':[],'j':[],'p':[]}
            for x in range(len(lines)):
                if x>1:
                    ls=lines[x].split("\t")
                    out["i"]+=[int(ls[0])]
                    out["j"]+=[int(ls[1])]
                    out["p"]+=[float(ls[2])]
        return self.__cast_pairing_prob(out)

    def __cast_pairing_prob(self, prob:dict)->list:
        """Computes the pairing probability list.

        Args:
            prob (dict): Ouput of RNAstructure.
            index (int): index of the row that you want the pairing probabilities of.

        Returns:
            list: pairing probability, under the form of a list of probabilities
        """
        # Create local dataframe to play here
        df_loc = pd.DataFrame(prob)
        df_loc = pd.concat((df_loc, df_loc.rename(columns={'i':'j','j':'i'})))
        # Group the probabilities by i and get an ordered list of the pairing probability
        g = df_loc.groupby('i')['p'].agg(lambda row: [pow(10,-float(r)) for r in row])
        g.index = g.index.astype(int)
        g = g.sort_index()
        g['sum_log_p'] = g.apply(lambda row: sum(row))
        # round to 4 
        g['sum_log_p'] = g['sum_log_p'].apply(lambda x: round(x,4))
        return list(g['sum_log_p'])
    
    def run_sequence_only(self, sequence):
        os.makedirs(os.path.join(self.config['temp_folder']), exist_ok=True)
        temp_prefix = os.path.join(self.config['temp_folder'], sequence)
        self.make_files(temp_prefix)
        if not os.path.isfile(temp_prefix+'.fasta'):
            self.create_fasta_file(temp_prefix, sequence)
            self.predict_construct()
        out = {}
        out['deltaG'], out['structure'] = self.extract_deltaG_struct()
        return out

    def run(self, mh, sample, sequence_only=False):
        if sequence_only:
            return self.run_sequence_only(mh['sequence'])
        out = {}
        temp_folder = util.make_folder(os.path.join(self.config['temp_folder'], str(sample)))
        temp_prefix = os.path.join(temp_folder, mh['construct'])
        if not os.path.exists(temp_prefix):
            os.makedirs(temp_prefix)
        temp_prefix = os.path.join(temp_folder, mh['section'])
        self.generate_normalized_mut_rates(temp_prefix, mh.info_bases, mh.mut_bases)
        for temperature, temperature_suf in {False:'', True:'_T'}.items():
            if temperature and not self.config['temperature']:
                continue
            this_sequence = mh['sequence']
            for dms, dms_suf in {False:'', True:'_DMS'}.items():
                if dms and min(mh['info_bases']) == 0 or dms and not self.config['dms']:
                    continue
                suffix = dms_suf+temperature_suf
                self.make_files(f"{temp_prefix}{suffix}")
                self.create_fasta_file(mh['construct'], this_sequence)
                self.predict_construct(use_dms = dms, dms_file=temp_prefix+"_DMS_signal.txt", use_temperature=temperature, temperature_k=mh.temperature_k)
                out['deltaG'+suffix], out['structure'+suffix] = self.extract_deltaG_struct()
                if not dms and self.config['partition']:
                    out['deltaG_ens'+suffix] = self.predict_ensemble_energy()
                if not dms  and not temperature and self.config['probability']:
                    out['mut_probability'+suffix] = self.predict_mut_probability(use_temperature=temperature, temperature_k=mh.temperature_k)

        return dict(sorted(out.items()))


def add_rnastructure_predictions(config, sample, mh, verbose=False):
    rna_pred = {}
    # Create the RNAstructure object
    rna = RNAstructure(config)

    return rna.run(mh, sample)





import shlex
import subprocess,os

from typing import Any, List


# Commands for external applications

BOWTIE2_CMD = "bowtie2"
BOWTIE2_BUILD_CMD = "bowtie2-build"
CUTADAPT_CMD = "cutadapt"
FASTQC_CMD = "fastqc"
SAMTOOLS_CMD = "samtools"


# Command utility functions

def run_cmd(args: List[Any], shell: bool = False):
    args_str = tuple(map(str, args))
    # FIXME: decide on either subprocess or os.system
    #subprocess.run(args_str, check=True, shell=shell)
    #cmd = shlex.join(args)
    cmd = ' '.join(args_str)
    excode = os.system(cmd)
    if excode:
        raise OSError(f"Command '{cmd}' returned exit code {excode}")
    return cmd



from typing import Dict, Set

# Constants

BASES = b"ACGT"
COMPS = b"TGCA"
RBASE = b"ACGU"
RCOMP = b"UGCA"
BASEN = b"N"
BASES_SET = set(BASES)
BASEN_SET = set(BASES + BASEN)



def get_diffs(seq1, seq2):
    if len(seq1) != len(seq2):
        raise ValueError("Sequences were different lengths: "
                         f"'{seq1}' and '{seq2}'")
    diffs = [i for i, (x1, x2) in enumerate(zip(seq1, seq2)) if x1 != x2]
    return diffs


class Seq(bytes):
    __slots__ = []

    alph = b""
    comp = b""
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def __init__(self, seq: bytes):
        self.validate_seq(seq)
        super().__init__()
    
    @classmethod
    def validate_seq(cls, seq):
        if not seq:
            raise ValueError("seq is empty")
        if set(seq) - cls.alphaset:
            raise ValueError(f"Invalid characters in seq: '{seq.decode()}'")

    @property
    def rc(self):
        return self.__class__(self[::-1].translate(self.trans))

    def __getitem__(self, item):
        return self.__class__(super().__getitem__(item))

    def __str__(self):
        return self.decode()


class DNA(Seq):
    alph = BASES
    comp = COMPS
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def tr(self):
        """
        Transcribe DNA into RNA.
        """
        return RNA(self.replace(b"T", b"U"))


class RNA(Seq):
    alph = RBASE
    comp = RCOMP
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def rt(self):
        """
        Reverse transcribe RNA into DNA.
        """
        return DNA(self.replace(b"U", b"T"))


class FastaIO(object):
    __slots__ = ["_path", "_refs"]

    recsym = b">"
    deftrunc = len(recsym)

    def __init__(self, path: str):
        self._path = path


class FastaParser(FastaIO):
    def __init__(self, path: str):
        super().__init__(path)
        self._refs: Set[str] = set()

    @classmethod
    def _parse_fasta_record(cls, fasta, line: bytes):
        if not line.startswith(cls.recsym):
            raise ValueError("FASTA definition line does not start with "
                             f"'{cls.recsym.decode()}'")
        name = line.split()[0][cls.deftrunc:].decode()
        seq = bytearray()
        while (line := fasta.readline()) and not line.startswith(cls.recsym):
            seq.extend(line.rstrip())
        seq = DNA(bytes(seq))
        return line, name, seq

    def parse(self):
        with open(self._path, "rb") as f:
            line = f.readline()
            while line:
                line, name, seq = self._parse_fasta_record(f, line)
                if name in self._refs:
                    raise ValueError(
                        f"Duplicate entry in {self._path}: '{name}'")
                self._refs.add(name)
                yield name, seq


class FastaWriter(FastaIO):
    def __init__(self, path: str, refs: Dict[str, DNA]):
        super().__init__(path)
        self._refs = refs
    
    def write(self):
        with open(self._path, "wb") as f:
            for ref, seq in self._refs.items():
                f.write(b"".join(
                    [self.recsym, ref.encode(), b"\n", seq, b"\n"]
                ))
                

def parse_fasta(path: str):
    return FastaParser(path).parse()



import os
import itertools
import unittest
from unittest import TestCase

from dreem.util.util import *
from dreem.vector.vector import *



class TestConsensusByte(TestCase):
    pass



class TestSamFlag(TestCase):
    flags = ("paired", "proper", "unmap", "munmap", "rev", "mrev",
             "first", "second", "secondary", "qcfail", "dup", "supp")

    bits = len(flags)

    # attributes

    def test_flag_attrs(self):
        sf = SamFlag(0)
        self.assertListEqual(sf.__slots__, list(self.flags))
        self.assertEqual(sf.MAX_FLAG, 4095)

    # valid flags

    def test_flag_0(self):
        sf = SamFlag(0)
        self.assertFalse(any(sf.__getattribute__(flag) for flag in self.flags))

    def test_flag_99(self):
        sf = SamFlag(99)
        self.assertTrue(all([sf.paired, sf.proper, sf.mrev, sf.first]))
        self.assertFalse(any([sf.unmap, sf.munmap, sf.rev, sf.second,
                              sf.secondary, sf.qcfail, sf.dup, sf.supp]))

    def test_flag_4095(self):
        sf = SamFlag(4095)
        self.assertTrue(all(sf.__getattribute__(flag) for flag in self.flags))

    def test_flags_all_valid(self):
        for flag, flags in enumerate(itertools.product([False, True],
                                                       repeat=self.bits)):
            sf = SamFlag(flag)
            sf_flags = list(map(sf.__getattribute__, self.flags))
            real_flags = list(reversed(flags))
            self.assertListEqual(sf_flags, real_flags)

    # invalid flags

    def test_flag_4096(self):
        with self.assertRaises(ValueError):
            SamFlag(4096)

    def test_flag_neg1(self):
        with self.assertRaises(ValueError):
            SamFlag(-1)

    def test_flag_float(self):
        with self.assertRaises(TypeError):
            SamFlag(99.0)

    def test_flag_str(self):
        with self.assertRaises(TypeError):
            SamFlag("99")

    def test_flag_bytes(self):
        with self.assertRaises(TypeError):
            SamFlag(b"99")


class TestSamRead(TestCase):
    # valid SAM lines

    def test_read_end5(self):
        line = b"FS10000136:97:BPN80019-0831:1:1101:6870:1040	99	end5	1	22	31M65S	=	1	-96	CAGCACTCAGAGCTAATACGACTCACTATAGATAATTGTGTACAAAGTAGAGATGTATCCAATTATGTGACTACCTTTGTGTAATAAAAATTTGTT	,FFFFFF:FF,:FFFFF:::,FF,F,FFF,:F:F,FF:F,F,:F,::F,FF,:,:FFF:F,FFF:FFF:,F,F::FF::FFF,F:FF::,FFF,:F	MD:Z:31	XG:i:0	NM:i:0	XM:i:0	XN:i:0	XO:i:0	AS:i:62	YS:i:62	YT:Z:CP"
        read = SamRead(line)
        sf99 = SamFlag(99)
        self.assertEqual(read.qname, b"FS10000136:97:BPN80019-0831:1:1101:6870:1040")
        self.assertListEqual(
            list(map(read.flag.__getattribute__, TestSamFlag.flags)),
            list(map(sf99.__getattribute__, TestSamFlag.flags)))
        self.assertEqual(read.rname, b"end5")
        self.assertEqual(read.pos, 1)
        self.assertEqual(read.cigar, b"31M65S")
        self.assertEqual(read.seq, b"CAGCACTCAGAGCTAATACGACTCACTATAGATAATTGTGTACAAAGTAGAGATGTATCCAATTATGTGACTACCTTTGTGTAATAAAAATTTGTT")
        self.assertEqual(len(read), 96)

    # invalid SAM lines

    def test_seq_qual_different_lengths(self):
        with self.assertRaises(ValueError):
            SamRead(b"Q	0	R	1	100	4=	*	*	4	ACGT	III")

    def test_read_empty(self):
        with self.assertRaises(ValueError):
            SamRead(b"")

    def test_read_str(self):
        with self.assertRaises(TypeError):
            line = "FS10000136:97:BPN80019-0831:1:1101:6870:1040	99	end5	1	22	31M65S	=	1	-96	CAGCACTCAGAGCTAATACGACTCACTATAGATAATTGTGTACAAAGTAGAGATGTATCCAATTATGTGACTACCTTTGTGTAATAAAAATTTGTT	,FFFFFF:FF,:FFFFF:::,FF,F,FFF,:F:F,FF:F,F,:F,::F,FF,:,:FFF:F,FFF:FFF:,F,F::FF::FFF,F:FF::,FFF,:F	MD:Z:31	XG:i:0	NM:i:0	XM:i:0	XN:i:0	XO:i:0	AS:i:62	YS:i:62	YT:Z:CP"
            SamRead(line)


class TestParseCigar(TestCase):
    # valid CIGAR strings

    def test_cigar_1op(self):
        parsed = list(parse_cigar(b"300="))
        self.assertListEqual(parsed, [(b"=", 300)])

    def test_cigar_5ops(self):
        parsed = list(parse_cigar(b"5I1X65=1D8S"))
        self.assertListEqual(parsed,
            [(b"I", 5), (b"X", 1), (b"=", 65), (b"D", 1), (b"S", 8)])

    # invalid CIGAR strings

    def test_cigar_wrong_order(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"=300"))

    def test_cigar_double_op(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"300=="))

    def test_cigar_invalid_op(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"300x"))

    def test_cigar_missing_op(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"300"))

    def test_cigar_neg_length(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"-300="))

    def test_cigar_missing_length(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"="))

    def test_cigar_zero_length(self):
        with self.assertRaises(ValueError):
            list(parse_cigar(b"0="))


class TestVectorizeReadMatchSub(TestCase):
    def test_valid_span_match_eq(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	4=	*	*	4	ACGT	IIII"
        expect = MATCH*4
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)
    
    def test_valid_span_match_m(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	4M	*	*	4	ACGT	IIII"
        expect = MATCH*4
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)


class TestVectorizeReadOneDel(TestCase):
    def test_1del_5prm_match_match(self):
        ref = b"ACCG"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	ACG	III"
        expect = MATCH + MADEL*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_match_match(self):
        ref = b"TGGC"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	TGC	III"
        expect = MATCH + MADEL*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_5prm_match_match_low_qual(self):
        ref = b"ACCG"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	ACG	I!I"
        DSM_C = (MADEL[0]|(SUB_N[0]^SUB_C[0])).to_bytes()
        expect = MATCH + DSM_C*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_match_match_low_qual(self):
        ref = b"TGGC"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	TGC	I!I"
        DSM_G = (MADEL[0]|(SUB_N[0]^SUB_G[0])).to_bytes()
        expect = MATCH + DSM_G*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_5prm_mismatch_mismatch_same(self):
        ref = b"ACCG"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	ATG	III"
        DEL_T = (DELET[0]|SUB_T[0]).to_bytes()
        expect = MATCH + DEL_T*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_mismatch_mismatch_same(self):
        ref = b"TGGC"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	TAC	III"
        DEL_A = (DELET[0]|SUB_A[0]).to_bytes()
        expect = MATCH + DEL_A*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_5prm_mismatch_mismatch_diff(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	ATT	III"
        DEL_T = (DELET[0]|SUB_T[0]).to_bytes()
        expect = MATCH + DEL_T*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_mismatch_mismatch_diff(self):
        ref = b"TGCA"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	TAA	III"
        DEL_A = (DELET[0]|SUB_A[0]).to_bytes()
        expect = MATCH + DEL_A*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_5prm_match_mismatch(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	AGT	III"
        expect = MATCH + DELET + MATCH*2
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_match_mismatch(self):
        ref = b"TCGA"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	TCA	III"
        expect = MATCH*2 + DELET + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_5prm_match_mismatch_low_qual(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	AGT	I!I"
        DSM_C = (MADEL[0]|(SUB_N[0]^SUB_C[0])).to_bytes()
        DSM_G = (MADEL[0]|(SUB_N[0]^SUB_G[0])).to_bytes()
        expect = MATCH + DSM_C + DSM_G + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_match_mismatch_low_qual(self):
        ref = b"CATG"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	CAG	I!I"
        DSM_A = (MADEL[0]|(SUB_N[0]^SUB_A[0])).to_bytes()
        DSM_T = (MADEL[0]|(SUB_N[0]^SUB_T[0])).to_bytes()
        expect = MATCH + DSM_A + DSM_T + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_5prm_mismatch_match(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1D2M	*	*	4	ACT	III"
        expect = MATCH + DELET + SUB_C + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1del_3prm_mismatch_match(self):
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1D1M	*	*	4	AGT	III"
        expect = MATCH + SUB_G + DELET + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)


class TestVectorizeReadMultiDels(TestCase):
    def test_2of2dels(self):
        ref = b"ACCG"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M2D1M	*	*	4	AG	II"
        expect = MATCH + DELET*2 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2of3dels_1(self):
        ref = b"ACCCG"
        first, last = 1, 5
        line = b"Q	0	R	1	100	1M2D2M	*	*	4	ACG	III"
        expect = MATCH + MADEL*3 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2of3dels_2(self):
        ref = b"ACCCG"
        first, last = 1, 5
        line = b"Q	0	R	1	100	1M1D1M1D1M	*	*	4	ACG	III"
        expect = MATCH + MADEL*3 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2of3dels_3(self):
        ref = b"ACCCG"
        first, last = 1, 5
        line = b"Q	0	R	1	100	2M2D1M	*	*	4	ACG	III"
        expect = MATCH + MADEL*3 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2of4dels_1(self):
        ref = b"ACCGGT"
        first, last = 1, 6
        line = b"Q	0	R	1	100	1M1D2M1D1M	*	*	4	ACGT	IIII"
        expect = MATCH + MADEL*4 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2of4dels_2(self):
        ref = b"ACCGGT"
        first, last = 1, 6
        line = b"Q	0	R	1	100	2M2D2M	*	*	4	ACGT	IIII"
        expect = MATCH + MADEL*4 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2of6dels(self):
        ref = b"ACCCCCCG"
        first, last = 1, 8
        line = b"Q	0	R	1	100	3M2D3M	*	*	4	ACCCCG	IIIIII"
        expect = MATCH + MADEL*6 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)
    
    def test_tunnel_2_5prm(self):
        ref = b"ACGCGCGT"
        first, last = 1, 8
        line = b"Q	0	R	1	100	1M2D5M	*	*	4	ACGCGT	IIIIII"
        expect = MATCH + MADEL*6 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_tunnel_2_mid(self):
        ref = b"ACGCGCGT"
        first, last = 1, 8
        line = b"Q	0	R	1	100	3M2D3M	*	*	4	ACGCGT	IIIIII"
        expect = MATCH + MADEL*6 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_tunnel_2_3prm(self):
        ref = b"ACGCGCGT"
        first, last = 1, 8
        line = b"Q	0	R	1	100	5M2D1M	*	*	4	ACGCGT	IIIIII"
        expect = MATCH + MADEL*6 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_tunnel_3_5prm(self):
        ref = b"TCAGCAGCAT"
        first, last = 1, 10
        line = b"Q	0	R	1	100	1M3D6M	*	*	4	TCAGCAT	IIIIIII"
        expect = MATCH + MADEL*8 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_tunnel_3_mid(self):
        ref = b"TCAGCAGCAT"
        first, last = 1, 10
        line = b"Q	0	R	1	100	2M3D5M	*	*	4	TCAGCAT	IIIIIII"
        expect = MATCH + MADEL*8 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_tunnel_3_3prm(self):
        ref = b"TCAGCAGCAT"
        first, last = 1, 10
        line = b"Q	0	R	1	100	6M3D1M	*	*	4	TCAGCAT	IIIIIII"
        expect = MATCH + MADEL*8 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)


class TestVectorizeReadOneIns(TestCase):
    def test_1ins(self):
        ref = b"ACTT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1I2M	*	*	4	ACGTT	IIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = MATCH + MIN_5 + MIN_3 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_5prm_match_match(self):
        ref = b"ACT"
        first, last = 1, 3
        line = b"Q	0	R	1	100	1M1I2M	*	*	4	ACCT	IIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MIN_5 + MIN_B + MIN_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_3prm_match_match(self):
        ref = b"ACT"
        first, last = 1, 3
        line = b"Q	0	R	1	100	2M1I1M	*	*	4	ACCT	IIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MIN_5 + MIN_B + MIN_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)
    
    def test_1ins_5prm_match_match_match(self):
        ref = b"ACCT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1I3M	*	*	5	ACCCT	IIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MIN_5 + MIN_B*2 + MIN_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_center_match_match_match(self):
        ref = b"ACCT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1I2M	*	*	5	ACCCT	IIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MIN_5 + MIN_B*2 + MIN_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_3prm_match_match_match(self):
        ref = b"ACCT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	3M1I1M	*	*	5	ACCCT	IIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MIN_5 + MIN_B*2 + MIN_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_5prm_match_lowq(self):
        ref = b"ACTA"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1I3M	*	*	4	ACGTA	II!II"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_C[0])).to_bytes()
        POS_3 = (MATCH[0]|INS_3[0]|INS_5[0]|(ANY_N[0]^SUB_T[0])).to_bytes()
        POS_4 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3 + POS_4
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_mid_match_lowq(self):
        ref = b"ACTA"
        first, last = 1, 4
        line = b"Q	0	R	1	100	2M1I2M	*	*	4	ACGTA	II!II"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_C[0])).to_bytes()
        POS_3 = (MATCH[0]|INS_3[0]|INS_5[0]|(ANY_N[0]^SUB_T[0])).to_bytes()
        POS_4 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3 + POS_4
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_3prm_match_lowq(self):
        ref = b"ACTA"
        first, last = 1, 4
        line = b"Q	0	R	1	100	3M1I1M	*	*	4	ACGTA	II!II"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_C[0])).to_bytes()
        POS_3 = (MATCH[0]|INS_3[0]|INS_5[0]|(ANY_N[0]^SUB_T[0])).to_bytes()
        POS_4 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3 + POS_4
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_5prm_lowq_lowq(self):
        """
        Possible alignments:
        Ref:    A-CATA    AC-ATA    ACA-TA    ACAT-A
        Read:   ACNNTA    ACNNTA    ACNNTA    ACNNTA
        """
        ref = b"ACATA"
        first, last = 1, 5
        line = b"Q	0	R	1	100	2M1I3M	*	*	6	ACGATA	II!!II"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_C[0])).to_bytes()
        POS_3 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_A[0])).to_bytes()
        POS_4 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_T[0])).to_bytes()
        POS_5 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3 + POS_4 + POS_5
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)
    
    def test_1ins_3prm_lowq_lowq(self):
        """
        Possible alignments:
        Ref:    A-CGTA    AC-GTA    ACG-TA    ACGT-A
        Read:   ACNNTA    ACNNTA    ACNNTA    ACNNTA
        """
        ref = b"ACGTA"
        first, last = 1, 5
        line = b"Q	0	R	1	100	2M1I3M	*	*	6	ACGATA	II!!II"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_C[0])).to_bytes()
        POS_3 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_G[0])).to_bytes()
        POS_4 = (MATCH[0]|INS_5[0]|INS_3[0]|(ANY_N[0]^SUB_T[0])).to_bytes()
        POS_5 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3 + POS_4 + POS_5
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_5prm_mismatch_match(self):
        """
        Possible alignments:
        Read:   CAGT
        Ref:    C-AT
        """
        ref = b"CAT"
        first, last = 1, 3
        line = b"Q	0	R	1	100	1M1I2M	*	*	4	CAGT	IIII"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (SUB_G[0]|INS_3[0]).to_bytes()
        POS_3 = MATCH
        expect = POS_1 + POS_2 + POS_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)
    
    def test_1ins_3prm_mismatch_match(self):
        """
        Possible alignments:
        Read:   CAGT
        Ref:    CG-T
        """
        ref = b"CGT"
        first, last = 1, 3
        line = b"Q	0	R	1	100	2M1I1M	*	*	4	CAGT	IIII"
        POS_1 = MATCH
        POS_2 = (SUB_A[0]|INS_5[0]).to_bytes()
        POS_3 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_5prm_mismatch_mismatch(self):
        """
        Possible alignments:
        Read:   CATC    CATC
        Ref:    C-GC    CG-C
        """
        ref = b"CGC"
        first, last = 1, 3
        line = b"Q	0	R	1	100	1M1I2M	*	*	4	CATC	IIII"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (SUB_A[0]|SUB_T[0]|INS_5[0]|INS_3[0]).to_bytes()
        POS_3 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_3prm_mismatch_mismatch(self):
        """
        Possible alignments:
        Read:   CATC    CATC
        Ref:    C-GC    CG-C
        """
        ref = b"CGC"
        first, last = 1, 3
        line = b"Q	0	R	1	100	2M1I1M	*	*	4	CATC	IIII"
        POS_1 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_2 = (SUB_A[0]|SUB_T[0]|INS_5[0]|INS_3[0]).to_bytes()
        POS_3 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = POS_1 + POS_2 + POS_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_centered(self):
        """
        Possible alignments:
        Read:   AAATAAA
        Ref:    AAA-AAA
        """
        ref = b"AAAAAA"
        first, last = 1, 6
        line = b"Q	0	R	1	100	3M1I3M	*	*	4	AAATAAA	IIIIIII"
        POS_3 = (MATCH[0]|INS_5[0]).to_bytes()
        POS_4 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = MATCH*2 + POS_3 + POS_4 + MATCH*2
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_off_centered_5prm(self):
        """
        Possible alignments:
        Read:   AAATAAA    AAATAAA    AAATAAA
        Ref:    -AAAAAA    A-AAAAA    AA-AAAA
        """
        ref = b"AAAAAA"
        first, last = 1, 6
        line = b"Q	0	R	1	100	1M1I5M	*	*	4	AAATAAA	IIIIIII"
        MI_53 = (MATCH[0]|INS_5[0]|INS_3[0]).to_bytes()
        POS_3 = (SUB_T[0]|INS_3[0]).to_bytes()
        expect = MI_53*2 + POS_3 + MATCH*3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_1ins_off_centered_3prm(self):
        """
        Possible alignments:
        Read:   AAATAAA    AAATAAA    AAATAAA
        Ref:    AAAA-AA    AAAAA-A    AAAAAA-
        """
        ref = b"AAAAAA"
        first, last = 1, 6
        line = b"Q	0	R	1	100	5M1I1M	*	*	4	AAATAAA	IIIIIII"
        MI_53 = (MATCH[0]|INS_5[0]|INS_3[0]).to_bytes()
        POS_4 = (SUB_T[0]|INS_5[0]).to_bytes()
        expect = MATCH*3 + POS_4 + MI_53*2
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)


class TestVectorizeReadMultiInns(TestCase):
    def test_2ins_indep(self):
        """
        Possible alignments:
        Read:   ATCGAT
        Ref:    A-CG-T
        """
        ref = b"ACGT"
        first, last = 1, 4
        line = b"Q	0	R	1	100	1M1I2M1I1M	*	*	6	ATCGAT	IIIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        expect = MIN_5 + MIN_3 + MIN_5 + MIN_3
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2ins_5prm_ymdt230103(self):
        ref = b"GAGGATCTGCCTGTATCGTCCGGAG"
        first, last = 1, 25
        line = b"Q	0	R	1	100	7M1I14M1I4M	*	*	27	GAGGATCTTGCCTGTATCGTCCGGGAG	IIIIIIIIIIIIIIIIIIIIIIIIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MATCH*6 + MIN_5 + MIN_B + MIN_3 + MATCH*11 + MIN_5 + MIN_B*2 + MIN_3 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)

    def test_2ins_3prm_ymdt230103(self):
        ref = b"GAGGATCTGCCTGTATCGTCCGGAG"
        first, last = 1, 25
        line = b"Q	0	R	1	100	8M1I15M1I2M	*	*	27	GAGGATCTTGCCTGTATCGTCCGGGAG	IIIIIIIIIIIIIIIIIIIIIIIIIII"
        MIN_5 = (MATCH[0]|INS_5[0]).to_bytes()
        MIN_3 = (MATCH[0]|INS_3[0]).to_bytes()
        MIN_B = (MIN_5[0]|MIN_3[0]).to_bytes()
        expect = MATCH*6 + MIN_5 + MIN_B + MIN_3 + MATCH*11 + MIN_5 + MIN_B*2 + MIN_3 + MATCH
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)


class TestVectorizeRealData(TestCase):
    """
    Test on examples of reads from real DMS-MaPseq experiments.
    """
    def test_real_1(self):
        ref = b"TATGGCTGTAGTTGTGATCAACTCCGCGAACCCATGCTTCAGTCAGCTGATGCACAATCGTTTTTAAACGGGTTTGCGGTGTAAGTGCAGCCCGTCTTAC"
        first, last = 13400, 13499
        line = b"FS10000136:77:BPG61616-2310:1:1101:1170:1610	163	SARS2	13342	44	52=1D63=1X34=	=	13474	283	CCCTGTGGGTTTTACACTTAAAAACACAGTCTGTACCGTCTGCGGTATGTGGAAGGTTATGGCTGTAGTTGTGATCAACTCCGCGAACCCATGCTTCAGTCAGCTGATGCACAATTGTTTTTAAACGGGTTTGCGGTGTAAGTGCAGCCC	FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF	AS:i:147	XN:i:0	XM:i:1	XO:i:1	XG:i:1	NM:i:2	MD:Z:52^A63C34	YS:i:149	YT:Z:CP"
        expect = MATCH*58 + SUB_T + MATCH*34 + BLANK*7
        muts = vectorize_read(ref, first, last, SamRead(line))
        self.assertTrue(muts == expect)
    
    def test_real_2(self):
        ref = b"GTCGCTTCCAAGAAAAGGACGAAGATGACAATTTAATTGATTCTTACTTTGTAGTTAAGAGACACACTTTCTCTAACTACCAACATGAAGAAACAATTTATAATTTACTTAAGGATTGTCCAGCTGTTGCTAAACATGACTTCTTTAAGTTTAGAATAGACGGTGACATGGTACCACATATATCACGTCAACGTCTTACT"
        first, last = 13601, 13800
        line = b"FS10000136:77:BPG61616-2310:1:1101:14070:2080	83	SARS2	13621	44	44=1X37=1X33=1X3=1D1X4=1I2X1=1I2=2I1X4=1X1=1I4=3S	=	13469	-297	GAAGATGACAATTTAATTGATTCTTACTTTGTAGTTAAGAGACAGACTTTCTCTAACTACCAACATGAAGAAACAATTTATATTTTACTTAAGGATTGTCCAGCTGTTGCTAAACACGACACTTTGTCGCTTCCAAGAAAAGGACGAAG	FFF:FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF::FFFFFFFFFFFFFFFFFFFFFFFFFFFFF:FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF	AS:i:119	XN:i:0	XM:i:8	XO:i:5	XG:i:6	NM:i:14	MD:Z:44C37A33T3^T0T4A0A3T4T5	YS:i:146	YT:Z:CP"
        muts = vectorize_read(ref, first, last, SamRead(line))
        # FIXME: add assertion

class TestVectorizePair(TestCase):
    def test_valid_no_muts(self):
        ref = b"ACGT"
        first, last = 1, 4
        line1 = b"Q	0	R	1	100	4=	*	*	4	ACGT	IIII"
        line2 = line1
        expect = MATCH * 4
        muts = vectorize_pair(ref, first, last, SamRead(line1), SamRead(line2))
        self.assertTrue(muts == expect)


if __name__ == "__main__":
    unittest.main()



import pandas as pd
import yaml
import numpy as np

from dreem.aggregate.resources.get_attributes import read_sample_attributes

def get_samples_info(df_samples, sample, verbose= False):
    if verbose: print(f"Adding samples info for {sample}")

    # Sanity check
    # Keep only the columns that are in sample_attributes.yml
    df_samples = df_samples[df_samples['sample']==sample]

    assert len(df_samples) <= 1, f"{sample} has more than one line in samples.csv"
    assert len(df_samples) == 1, f"{sample} doesn't have a corresponding line in samples.csv"
    
    df_samples = df_samples.iloc[0]
    
    exp_env = df_samples['exp_env']
    assert exp_env in ['in_vivo','in_vitro'], f"{exp_env} is not a valid value for exp_env. Should be in_vivo or in_vitro"

    # Load list of mandatory columns and check that they are in samples.csv and not empty for this sample 
    sample_attributes = read_sample_attributes()
    for mand in sample_attributes['mandatory']['all'] + sample_attributes['mandatory'][exp_env]:
        assert mand in list(df_samples.index), f"{mand} is not in samples.csv"
        assert not df_samples[mand] == np.nan, f"{mand} is empty in samples.csv for sample {sample}"
        
    return df_samples.to_dict()

def get_library_info(df_library, construct, verbose= False):
    if verbose: print(f"Adding library info for {construct}")

    # Sanity check
    df_library = df_library[df_library['construct']==construct]

    df_library.drop(columns = [c for c in df_library.columns if c in ['section_start','section_end','section','construct']], inplace=True)
    
    return df_library.iloc[0].to_dict()






import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

import dreem.util.util as util
import pandas as pd

from dreem import demultiplexing

from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
from dreem.test.sample_profile import sample_profile, constructs, sections

module = 'demultiplexing'
sample_name = 'test_set_1'

module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output = os.path.join(output_dir, module)

inputs = ['fastq','library', 'fasta']
outputs = ['demultiplexed_fastq']

# ### Create test files for `test set 1`
def test_make_files():
    if not os.path.exists(os.path.join(test_files_dir, 'input', module)):
        os.makedirs(os.path.join(test_files_dir, 'input', module))
    files_generator.generate_files(sample_profile, module, inputs, outputs, test_files_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, outputs, prediction_dir, sample_name)

# ## Test code for convolution algorithm for demultiplexing

def test_run():
    for sample in os.listdir(module_input):
        demultiplexing.run(
            fastq1 = '{}/{}_R1.fastq'.format(os.path.join(module_input,sample),sample),
            fastq2 = '{}/{}_R2.fastq'.format(os.path.join(module_input,sample),sample),
            fasta= '{}/reference.fasta'.format(os.path.join(module_input,sample),sample),
            library = '{}/library.csv'.format(os.path.join(module_input,sample)),
            out_dir = os.path.join(module_output,sample)
            )

def test_output_exists():        
    files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, sample_name)

def test_all_files_are_equal():
    for sample in os.listdir(module_input):
        for pred, out in zip(os.listdir(os.path.join(module_expected,sample)), [f for f in os.listdir(os.path.join(module_output,sample)) if not f.startswith('lost_reads') and f.endswith('.fastq')]):
            assert pred == out, 'The expected output and the output files are not the same'
            expected = util.fastq_to_df(os.path.join(module_expected,sample,pred))
            expected['from'] = 'expected'
            output = util.fastq_to_df(os.path.join(module_output,sample,out))
            output['from'] = 'output'
            both = pd.concat([expected,output],axis=0, ignore_index=True)
            for idx, g in both.groupby('construct'):
                if len(g) < 2:
                    assert g['from'].iloc[0] == 'expected', 'The output file is missing the construct {} for file {}/{}'.format(idx,sample,out)
                    assert g['from'].iloc[0] == 'output', 'The output file didn\'t filter out the construct {} for file {}/{}'.format(idx,sample,out)
            for idx, g in both.groupby('construct'):
                for c in both.columns:
                    if c != 'construct' and c != 'from':
                        assert g[c].unique().shape[0] == 1, 'The output file is not the same as the expected output for sample {} and construct {}'.format(sample,idx)
                        